<?php

namespace FolderShare;

use FolderShare\FolderShareConnect;
use FolderShare\FolderShareFormat;

/**
 * Implements Linux/BSD/macOS-like commands for access to FolderShare content.
 *
 * This class provides a variety of utility methods that implement
 * commands similar to standard Linux/BSD/macOS file system commands,
 * such as "ls", "stat", "mkdir", "cp", etc.
 */
class FolderShareCommands {

  /*----------------------------------------------------------------------
   *
   * Constants.
   *
   *----------------------------------------------------------------------*/

  /**
   * The default wait interval when waiting for items to become enabled.
   *
   * @var int
   */
  const DEFAULT_WAIT_UNTIL_ENABLED_TIME = 5;

  /**
   * The anonymous login.
   *
   * Technically, the name of the anonymous account can be changed by
   * the server's site admin. This value is only used when that changed
   * value is unknown.
   *
   * @var string
   */
  const ANONYMOUS = 'anonymous';

  /*----------------------------------------------------------------------
   *
   * Available commands.
   *
   *----------------------------------------------------------------------*/

  /**
   * Describes commands and their options.
   *
   * The associative array's keys are command names available for use on the
   * command line (e.g. "ls" and "stat"). For each command, an array of values
   * describes the command, including its help text, a list of supported
   * flags, its return formats, synonyms, and finally the name of the function
   * to invoke to execute the command.
   *
   * Command help:
   *
   * - 'synopsis' = a one-line note on the command name and options, similar to
   *   the synopsis line of a traditional Linux/macOS/BSD man page.
   *
   * - 'description' = a multiparagraph description of the command.
   *
   * - 'flags' = an associative array that lists no-argument flags for the
   *   command. For each flag, the array key names the flag, including leading
   *   dashes ("-" or "--" or both). The key's value is an associative array
   *   with keys for:
   *   - 'brief' = a one-line description of the flag.
   *   - 'synonyms' = an array of equivalent flag names. Defaults to empty.
   *
   * - 'synonyms' = an array of alternate names for the command. Defaults to
   *   empty.
   *
   * Command return format keys:
   *
   * - 'formats' = an array of return format names supported by the command.
   *   Known values are: 'full', 'keyvalue', 'text', and 'linux'.
   *
   * - 'defaultFormat' = the default format.
   *
   * Execution keys:
   *
   * - 'function' = the name of the function to invoke for the command.
   *
   * @var array
   */
  const COMMANDS = [
    //
    // Local state.
    // ------------
    //
    // Print user name.
    //
    'whoami' => [
      'synopsis' => 'whoami',
      'brief'    => 'Show the current user name.',
      'description' => <<<'EOS'
Show the user name of the current user.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
      ],
      'synonyms' => [
        'id',
      ],
      'formats' => [],
      'defaultFormat' => '',
      'function' => 'FolderShareCommands::whoami',
    ],

    //
    // Print host name.
    //
    'hostname' => [
      'synopsis' => 'hostname',
      'brief'    => 'Show the current host name.',
      'description' => <<<'EOS'
Show the host name of the current server connection.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
      ],
      'formats' => [],
      'defaultFormat' => '',
      'function' => 'FolderSharecommands::hostname',
    ],

    //
    // HTTP GET operations supported.
    // ------------------------------
    //
    // GET version numbers.
    //
    'version' => [
      'synopsis' => 'version [OPTIONS]',
      'brief' => 'Show version numbers.',
      'description' => <<<'EOS'
Show version numbers for client and server software.

If a host name is provided, the host is contacted and its software version
numbers reported along with those for the client software. But if no host
name is provided, only client software version numbers are reported.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
      ],
      'formats' => [
        'full',
        'keyvalue',
        'text',
      ],
      'defaultFormat' => 'text',
      'function' => 'FolderShareCommands::version',
    ],

    //
    // GET server configuration.
    //
    'config' => [
      'synopsis' => 'config [OPTIONS]',
      'brief' => 'Show the host administrator configuration.',
      'description' => <<<'EOS'
Summarize the administrator configuration settings for the host.

Reported configuration values include:
  - A list of HTTP verbs supported by the host.
  - The serialization formats supported for each HTTP verb.
  - The authentication providers supported for each HTTP verb.
  - The maximum file upload size and number.
  - The file name extensions supported, if restrictions are enabled.

This information is primiarily of interest to host administrators and to those
developing client-server software. The reported values cannot be altered
through this client software. Administrators may adjust their host's
configuration using the web interface to the REST and FolderShare modules.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
      ],
      'synonyms' => [
        'configuration',
      ],
      'formats' => [
        'full',
        'keyvalue',
        'text',
      ],
      'defaultFormat' => 'text',
      'function' => 'FolderShareCommands::config',
    ],

    //
    // GET user's usage.
    //
    'usage' => [
      'synopsis' => 'usage [OPTIONS]',
      'brief' => 'Show the user\'s storage use.',
      'description' => <<<'EOS'
Report the number of top-level folders, subfolders, files, and bytes in use
by the user on the host.

The user must have permission to create and modify remote files and folders.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
      ],
      'formats' => [
        'full',
        'keyvalue',
        'text',
      ],
      'defaultFormat' => 'text',
      'function' => 'FolderShareCommands::usage',
    ],

    //
    // GET file or folder status.
    //
    'stat' => [
      'synopsis' => 'stat [OPTIONS] REMOTE_PATHS...',
      'brief' => 'Show the status of files and folders.',
      'description' => <<<'EOS'
Show detailed status for remote files or folders.

Status information includes the item's name, remote folder path, owner, size,
MIME type, numeric entity ID, parent entity ID, top-level folder entity ID,
creation date, modified date, and whether the item is private, shared with
others, or shared with the public.

Paths to remote files and folders must start with '/'. Wild cards are not
supported.

The user must have permission to view the remote files and folders.

This command is similar to the 'stat' command on Linux, BSD, and macOS.
EOS
      ,
      'flags' => [
        '-t' => [
          'brief' => 'Show a terse one-line version of the information.',
          'synonyms' => [
            '--terse',
          ],
        ],
        '--help' => [
          'brief' => 'Show this help information.',
        ],
      ],
      'synonyms' => [
        'status',
      ],
      'formats' => [
        'full',
        'keyvalue',
        'linux',
        'text',
      ],
      'defaultFormat' => 'linux',
      'function' => 'FolderSharecommands::stat',
    ],

    //
    // GET folder contents.
    //
    'ls' => [
      'synopsis' => 'ls [OPTIONS] REMOTE_PATHS...',
      'brief' => 'List files and folders.',
      'description' => <<<'EOS'
List remote files and folders.

The command supports a default short form and a long form with the '-l' option.
Both forms default to sorting items by name:
  - The short form lists item names in multiple columns.
  - The long form lists item names and attributes in a single-column list.

Paths to remote files and folders must start with '/'. Wild cards are not
supported.

The user must have permission to view the remote files and folders.

This command is similar to the 'ls' command on Linux, BSD, and macOS.
EOS
      ,
      'flags' => [
        '-1' => [
          'brief' => 'Force output to be one entry per line.',
        ],
        '-d' => [
          'brief' => 'List folders without recursion.',
        ],
        '-f' => [
          'brief' => 'Do not sort.',
        ],
        '-F' => [
          'brief' => 'Mark folders by adding a "/" suffix.',
        ],
        '-h' => [
          'brief' => 'Simplify numbers with byte suffixes (e.g. MB).',
        ],
        '-i' => [
          'brief' => 'Show the numeric entity ID for each item.',
        ],
        '-l' => [
          'brief' => 'Show a long form that lists more detail.',
        ],
        '-r' => [
          'brief' => 'Reverse the sort order.',
        ],
        '-R' => [
          'brief' => 'Recurse to list folder contents.',
        ],
        '-s' => [
          'brief' => 'Show the size, in bytes, for each item.',
        ],
        '-S' => [
          'brief' => 'Sort by size instead of name.',
        ],
        '-t' => [
          'brief' => 'Sort by modification time instead of name.',
        ],
        '--help' => [
          'brief' => 'Show this help information.',
        ],
      ],
      'synonyms' => [
        'list',
        'dir',
      ],
      'formats' => [
        'full',
        'keyvalue',
        'linux',
        'text',
      ],
      'defaultFormat' => 'linux',
      'function' => 'FolderShareCommands::ls',
    ],

    //
    // GET to download file.
    //
    'download' => [
      'synopsis' => 'download [OPTIONS] REMOTE_PATH [LOCAL_PATH]',
      'brief' => 'Download files and folders.',
      'description' => <<<'EOS'
Download a remote file or folder indicated to the local host.

The first path indicates a remote file or folder to download, while the second
path selects the local host location in which to save the downloaded item. If
the second path is omitted, the current directory is used and the new local
file has the same name as the remote file.

When downloading a remote folder, that folder and its contents are first
compressed into a ZIP archive and that archive is returned as a file saved onto
the local host. The ZIP archive is not automatically un-zipped.

By default, local files are silently overwritten by downloaded files. Add the
'-n' option to prevent overwriting. Downloaded files will be automatically
renamed with a unique sequence number starting at 2 (e.g. "file-2.txt").

The '-p' option sets downloaded file modification and access dates to match
those of the remote file.

The '-u' option updates downloaded files only if they are missing or the
remote file's modification date is newer than that of the local file. This
is typically used in combination with '-p' so that local and remote file
modification dates are comparable.

Paths to remote files and folders must start with '/'. Wild cards are not
supported.

The user must have permission to view the remote files and folders, and create
the local file.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
        '-n' => [
          'brief' => 'Do not overwrite an existing local file.',
          'synonyms' => [],
        ],
        '-p' => [
          'brief' => 'Preserve timestamps on downloaded items.',
          'synonyms' => [],
        ],
        '--sync' => [
          'brief' => 'Synchronize local and remote files (same as -pu).',
          'synonyms' => [],
        ],
        '-u' => [
          'brief' => 'Update by only downloading missing or newer files.',
          'synonyms' => [],
        ],
        '-v' => [
          'brief' => 'Show the name of each item as it is downloaded.',
          'synonyms' => [],
        ],
      ],
      'synonyms' => [
        'get',
      ],
      'formats' => [],
      'defaultFormat' => '',
      'function' => 'FolderShareCommands::download',
    ],

    //
    // GET to get specific attributes of a file or folder.
    //
    'xattr' => [
      'synopsis' => 'xattr [OPTIONS] REMOTE_PATHS',
      'brief' => 'Get extended attributes of files and folders.',
      'description' => <<<'EOS'
Get extended attributes of a remote file or folder.

The path selects a remote file or folder. The path must start with '/'.
Wild cards are not supported.

Extended attributes include whether a file or folder is disabled because
it is in use by another operation.

The user must have permission to view the remote file or folder.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
        '--disabled' => [
          'brief' => 'Return "true" if the item is disabled, "false" otherwise.',
          'synonyms' => [
            '--systemdisabled',
          ],
        ],
        '--hidden' => [
          'brief' => 'Return "true" if the item is hidden, "false" otherwise.',
          'synonyms' => [
            '--systemhidden',
          ],
        ],
        '--kind' => [
          'brief' => 'Return the item kind (e.g. "file", "folder", etc.',
          'synonyms' => [],
        ],
        '--mime' => [
          'brief' => 'Return the item MIME type.',
          'synonyms' => [
            '--mimetype',
          ],
        ],
        '--size' => [
          'brief' => 'Return the item size, in bytes.',
          'synonyms' => [],
        ],
        // TODO Add view and author grant UIDs.
      ],
      'synonyms' => [],
      'formats' => [],
      'defaultFormat' => '',
      'function' => 'FolderShareCommands::xattr',
    ],

    //
    // GET and wait to check when an item becomes enabled.
    //
    'wait' => [
      'synopsis' => 'wait REMOTE_PATHS',
      'brief' => 'Wait until remote files and folders are enabled.',
      'description' => <<<'EOS'
Synchronize with remote operations by waiting until an item is enabled.

The path selects a remote file or folder. The path must start with '/'.
Wild cards are not supported.

Operations that copy, move, or change ownership can take a long time and
require background processing. Files and folders involved are marked
disabled while processing is in progress. Meanwhile, the request returns
before the full operation is complete. This command waits until the
indicated file or folder becomes marked enabled.

The user must have permission to view the remote file or folder.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
        '-v' => [
          'brief' => 'Be verbose.',
        ],
      ],
      'synonyms' => [],
      'formats' => [],
      'defaultFormat' => '',
      'function' => 'FolderShareCommands::wait',
    ],

    //
    // HTTP DELETE operations supported.
    // ---------------------------------
    //
    // DELETE file or folder.
    //
    'rm' => [
      'synopsis' => 'rm [OPTIONS] REMOTE_PATHS...',
      'brief' => 'Remove files or folders.',
      'description' => <<<'EOS'
Remove remote files and and folders.

Use 'rm' without options to remove remote files. Use 'rm -d' to also remove
empty folders. Use 'rm -r' to recursively remove files and non-empty folders
and their contents, recursively.

Paths to remote files and folders must start with '/'. Wild cards are not
supported.

The user must have permission to remove the remote files and folders.

This command is similar to the 'rm' command on Linux, BSD, and macOS.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
        '-d' => [
          'brief' => 'Delete folders as well as files.',
          'synonyms' => [],
        ],
        '-f' => [
          'brief' => 'Force deletion to continue even after errors.',
          'synonyms' => [],
        ],
        '-r' => [
          'brief' => 'Recursively remove folders and their contents.',
          'synonyms' => [
            '-R',
          ],
        ],
        '-v' => [
          'brief' => 'Show the name of each item as it is deleted.',
          'synonyms' => [],
        ],
      ],
      'synonyms' => [
        'remove',
        'del',
        'delete',
      ],
      'formats' => [],
      'defaultFormat' => '',
      'function' => 'FolderShareCommands::rm',
    ],

    //
    // DELETE folder.
    //
    'rmdir' => [
      'synopsis' => 'rmdir [OPTIONS] REMOTE_PATHS...',
      'brief' => 'Remove empty folders.',
      'description' => <<<'EOS'
Remove remote empty folders.

Use 'rmdir' to remove remote empty folders. It is an error to attempt to
remove a non-empty folder.

To remove files and non-empty folders, use the 'rm -r' command instead.

Paths to remote files and folders must start with '/'. Wild cards are not
supported.

The user must have permission to remove the remote folders.

This command is similar to the 'rmdir' command on Linux, BSD, and macOS.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
        '-v' => [
          'brief' => 'Show the name of each item as it is deleted.',
          'synonyms' => [],
        ],
      ],
      'formats' => [],
      'defaultFormat' => '',
      'function' => 'FolderShareCommands::rmdir',
    ],

    //
    // HTTP POST operations supported.
    // -------------------------------
    //
    // POST new root folder or subfolder.
    //
    'mkdir' => [
      'synopsis' => 'mkdir [OPTIONS] REMOTE_PATHS...',
      'brief' => 'Make new folders.',
      'description' => <<<'EOS'
Make new remote empty folders.

Paths to remote files and folders must start with '/'. Wild cards are not
supported.

The user must have permission to create the remote folders.

This command is similar to the 'mkdir' command on Linux, BSD, and macOS.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
        '-p' => [
          'brief' => 'Create intermediate folders as required.',
          'synonyms' => [],
        ],
        '-v' => [
          'brief' => 'Show the name of each item as it is created.',
          'synonyms' => [],
        ],
      ],
      'formats' => [
        'full',
        'keyvalue',
      ],
      'defaultFormat' => 'keyvalue',
      'function' => 'FolderShareCommands::mkdir',
    ],

    //
    // POST to upload file.
    //
    'upload' => [
      'synopsis' => 'upload [OPTIONS] LOCAL_PATH REMOTE_PATH',
      'brief' => 'Upload files and folders.',
      'description' => <<<'EOS'
Upload a local file or folder to a remote folder.

The first path selects a local file or folder to upload, while the second
path indicates the remote location for the uploaded items.

Paths to remote files and folders must start with '/'. Wild cards are not
supported.

The user must have permission to view the local file and create the remote
files and folders.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
        '-n' => [
          'brief' => 'Do not overwrite an existing remote file.',
          'synonyms' => [],
        ],
        '-p' => [
          'brief' => 'Preserve timestamps on uploaded items.',
          'synonyms' => [],
        ],
        '--sync' => [
          'brief' => 'Synchronize local and remote files (same as -pu).',
          'synonyms' => [],
        ],
        '-u' => [
          'brief' => 'Update by only uploaded missing or newer files.',
          'synonyms' => [],
        ],
        '-v' => [
          'brief' => 'Show the name of each item as it is uploaded.',
          'synonyms' => [],
        ],
      ],
      'synonyms' => [
        'put',
      ],
      'formats' => [],
      'defaultFormat' => '',
      'function' => 'FolderShareCommands::upload',
    ],

    //
    // HTTP PATCH operations supported.
    // --------------------------------
    //
    // PATCH name and/or location.
    //
    'mv' => [
      'synopsis' => 'mv [OPTIONS] REMOTE_PATHS... REMOTE_DESTINATION_PATH',
      'brief' => 'Move files and folders.',
      'description' => <<<'EOS'
Move remote files or folders.

The command has two forms:
  - With two paths, the command moves the file or folder indicated by the
    first path into the destination selected by the second path.

  - With more than two paths, the command moves the files and folders indicated
    by all of the paths, except the last, into the destination folder selected
    by the last path.

With two paths, if the destination does not exist, but its parent folder does,
the moved item will be moved into the parent and renamed to use the name in
the destination path.

Paths to remote files and folders must start with '/'. Wild cards are not
supported.

The user must have permission to modify the remote folders.

This command is similar to the 'mv' command on Linux, BSD, and macOS.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
        '--wait' => [
          'brief' => 'Wait for move to complete.',
          'synonyms' => [
            '--wait',
          ],
        ],
        '-n' => [
          'brief' => 'Do not overwrite an existing file.',
          'synonyms' => [],
        ],
        '-v' => [
          'brief' => 'Show the name of each item as it is moved.',
          'synonyms' => [],
        ],
      ],
      'synonyms' => [
        'move',
      ],
      'formats' => [
        'full',
        'keyvalue',
      ],
      'defaultFormat' => 'keyvalue',
      'function' => 'FolderShareCommands::mv',
    ],

    //
    // PATCH to create a copy.
    //
    'cp' => [
      'synopsis' => 'cp [OPTIONS] REMOTE_PATHS... REMOTE_DESTINATION_PATH',
      'brief' => 'Copy files and folders.',
      'description' => <<<'EOS'
Copy remote files or folders.

The command has two forms:
  - With two paths, the command copies the file or folder indicated by the
    first path into the destination selected by the second path.

  - With more than two paths, the command copies the files and folders
    indicated by all of the paths, except the last, into the destination folder
    selected by the last path.

With two paths, if the destination does not exist, but its parent folder does,
the copied item will be copied into the parent and renamed to use the name in
the destination path.

Paths to remote files and folders must start with '/'. Wild cards are not
supported.

The user must have permission to create and modify the remote files and
folders.

This command is similar to the 'cp' command on Linux, BSD, and macOS.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
        '--wait' => [
          'brief' => 'Wait for copy to complete.',
          'synonyms' => [
            '--wait',
          ],
        ],
        '-n' => [
          'brief' => 'Do not overwrite an existing file.',
          'synonyms' => [],
        ],
        '-v' => [
          'brief' => 'Show the name of each item as it is copied.',
          'synonyms' => [],
        ],
      ],
      'synonyms' => [
        'copy',
      ],
      'formats' => [
        'full',
        'keyvalue',
      ],
      'defaultFormat' => 'keyvalue',
      'function' => 'FolderShareCommands::cp',
    ],

    //
    // PATCH to update fields.
    //
    'update' => [
      'synopsis' => 'update [OPTIONS] FIELD_NAME FIELD_VALUE REMOTE_PATHS...',
      'brief' => 'Update fields on files and folders.',
      'description' => <<<'EOS'
Update the value for a named field on remote files or folders.

Fields that may be updated include:
  - 'name' = the name of the remote file or folder (see also the 'mv' command).
  - 'description' = the text description of the remote file or folder.

Additional fields may be supported by third-party extensions or customizations
made on specific hosts.

Field values that are more than one word (such as a description's text) or that
include special characters (such as in HTML) should be surrounded by quotes.

Paths to remote files and folders must start with '/'. Wild cards are not
supported.

The user must have permission to modify the remote files and folders.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
        '-v' => [
          'brief' => 'Show the name of each item as it is changed.',
          'synonyms' => [],
        ],
      ],
      'formats' => [
        'full',
        'keyvalue',
        'text',
      ],
      'defaultFormat' => 'text',
      'function' => 'FolderShareCommands::update',
    ],

    //
    // PATCH to change ownership.
    //
    'chown' => [
      'synopsis' => 'chown [OPTIONS] USER REMOTE_PATHS...',
      'brief' => 'Change ownership on files and folders.',
      'description' => <<<'EOS'
Change the owner of a remote file or folder, and optionally its descendants.

The user argument gives a user account name or ID for the new owner. The
remainder of the arguments are remote paths to change.

Paths to remote files and folders must start with '/'. Wild cards are not
supported.

The user must have permission to modify the ownership of remote files and
folders.
EOS
      ,
      'flags' => [
        '--help' => [
          'brief' => 'Show this help information.',
        ],
        '-v' => [
          'brief' => 'Show the name of each item as it is changed.',
          'synonyms' => [],
        ],
        '-R' => [
          'brief' => 'Recursively change all descendents too.',
          'synonyms' => [],
        ],
      ],
      'formats' => [
        'full',
        'keyvalue',
      ],
      'defaultFormat' => 'keyvalue',
      'function' => 'FolderShareCommands::chown',
    ],
  ];

  /*--------------------------------------------------------------------
   *
   * Utilities.
   *
   *--------------------------------------------------------------------*/

  /**
   * Prints an error message to STDERR.
   *
   * @param string $commandName
   *   The name of the current command.
   * @param string $message
   *   The message to print.
   */
  private static function printError(
    string $commandName,
    string $message) {

    fprintf(STDERR, "%s: %s\n", $commandName, strip_tags($message));
  }

  /**
   * Prints an exception error message to STDERR.
   *
   * @param string $commandName
   *   The name of the current command.
   * @param \Exception $e
   *   The exception with a message to print.
   */
  private static function printException(
    string $commandName,
    \Exception $e) {

    fprintf(STDERR, "%s: %s\n", $commandName, strip_tags($e->getMessage()));
  }

  /**
   * Waits for all items to be enabled.
   *
   * This function enters an infinite loop. On each pass, it checks one
   * or more of the indicated remote paths to see if the corresponding
   * item is enabled. If any item is not enabled, the pass completes
   * and the process sleeps for several seconds before the next pass
   * through the loop.
   *
   * If all of the items are enabled, then all pending operations on them
   * are presumed to be complete and this function returns.
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param string $commandName
   *   The name of the command that called this function.
   * @param string[] $remoteUris
   *   The remote file or folder URIs to monitor.
   * @param bool $verbose
   *   (optional, default = FALSE) When TRUE, print a "." after each check
   *   of the items. When FALSE, print nothing.
   */
  public static function waitUntilEnabled(
    FolderShareConnect $server,
    string $commandName,
    array $remoteUris,
    bool $verbose = FALSE) {

    // Repeatedly get each entity and check if they are all enabled.
    // If any one of them is not, wait and repeat.
    while (TRUE) {
      $anyDisabled = FALSE;
      foreach ($remoteUris as $remoteUri) {
        try {
          $response = $server->getFileOrFolder($remoteUri, FALSE);
          if (isset($response['systemdisabled']) === TRUE &&
              isset($response['systemdisabled'][0]) === TRUE &&
              isset($response['systemdisabled'][0]['value']) === TRUE &&
              $response['systemdisabled'][0]['value'] === TRUE) {
            $anyDisabled = TRUE;
            break;
          }
        }
        catch (\Exception $e) {
          self::printException($commandName, $e);
          return FALSE;
        }
      }

      if ($anyDisabled === FALSE) {
        break;
      }

      if ($verbose === TRUE) {
        print(".");
        flush();
      }

      sleep(self::DEFAULT_WAIT_UNTIL_ENABLED_TIME);
    }
  }

  /*--------------------------------------------------------------------
   *
   * Connection commands.
   *
   *--------------------------------------------------------------------*/

  /**
   * Prints host configuration parameters.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | --help | Show help on command.                            |
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @see ::hostname()
   * @see ::version()
   */
  public static function config(
    FolderShareConnect $server,
    array $options) {

    //
    // Execute
    // -------
    // Report configuration.
    try {
      $response = $server->getServerConfiguration();

      switch ($options['format']) {
        case 'keyvalue':
        case 'text':
          print(FolderShareFormat::formatAsConfiguration($response));
          break;

        default:
        case 'full':
          print_r($response);
          break;
      }
    }
    catch (\Exception $e) {
      self::printException(reset($options['command']), $e);
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Prints the name of the current host.
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @see ::config()
   * @see ::version()
   */
  public static function hostname(
    FolderShareConnect $server,
    array $options) {

    printf("%s\n", $server->getHostNameAndPortNumber());
    return TRUE;
  }

  /**
   * Prints software version numbers.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | --help | Show help on command.                            |
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection. If the connection is NULL, only the client
   *   side's version information is returned.
   * @param array $options
   *   The command-line options array.
   * @param bool $doConnect
   *   (optional, default = TRUE) When TRUE, a request is sent to the server
   *   to get its version information, which is then merged with information
   *   about the client API and PHP version. When FALSE, no request is sent
   *   and the returned array only includes the client API and PHP version.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @see ::config()
   * @see ::hostname()
   */
  public static function version(
    FolderShareConnect $server,
    array $options,
    bool $doConnect = TRUE) {

    //
    // Execute
    // -------
    // Report configuration.
    try {
      $response = $server->getVersion($doConnect);

      if (is_array($response) === FALSE) {
        print_r($response);
      }
      else {
        if (isset($response['client']) === FALSE) {
          $response['client'] = [];
        }

        // Add an application entry to the front of the client version list.
        $response['client'] = array_merge(
          [
            'application' => [
              'name'      => NAME,
              'version'   => VERSION,
            ],
          ],
          $response['client']);

        if ($options['format'] === 'text') {
          print(FolderShareFormat::formatAsVersion($response));
        }
        else {
          print_r($response);
        }
      }
    }
    catch (\Exception $e) {
      self::printException(reset($options['command']), $e);
      return FALSE;
    }

    return TRUE;
  }

  /*--------------------------------------------------------------------
   *
   * User commands.
   *
   *--------------------------------------------------------------------*/

  /**
   * Prints the name of the current user.
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @see ::usage()
   */
  public static function whoami(
    FolderShareConnect $server,
    array $options) {

    $username = $server->getUserName();
    if (empty($username) === TRUE) {
      $username = self::ANONYMOUS;
    }

    printf("%s\n", $username);
    return TRUE;
  }

  /**
   * Prints file, folder, and storage usage for a user.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | --help | Show help on command.                            |
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @see ::whoami()
   */
  function usage(
    FolderShareConnect $server,
    array $options) {

    //
    // Execute
    // -------
    // Report usage.
    try {
      $response = $server->getUsage();

      switch ($options['format']) {
        case 'keyvalue':
        case 'text':
          print(FolderShareFormat::formatAsUsage($response));
          break;

        default:
        case 'full':
          print_r($response);
          break;
      }
    }
    catch (\Exception $e) {
      self::printException(reset($options['command']), $e);
      return FALSE;
    }

    return TRUE;
  }

  /*--------------------------------------------------------------------
   *
   * File and folder status commands.
   *
   *--------------------------------------------------------------------*/

  /**
   * Prints a list of files and folders.
   *
   * This command emulates the Linux/macOS/BSD "ls" command.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -1     | Force one-column output.                         |
   * | -d     | List directories without recursion.              |
   * | -f     | Do not sort.                                     |
   * | -F     | Add special symbols on files and folders.        |
   * | -h     | Simplify numbers with byte suffixes (e.g. MB).   |
   * | -i     | Include the entity ID.                           |
   * | -l     | Show a long listing format.                      |
   * | -r     | Reverse the sort order.                          |
   * | -R     | Recursively list directories.                    |
   * | -s     | Include the size of each item.                   |
   * | -S     | Sort by size.                                    |
   * | -t     | Sort by modification time.                       |
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @see ::chown()
   * @see ::stat()
   * @see ::update()
   * @see ::xattr()
   *
   * @internal
   * POSIX, Linux, macOS, and BSD "ls" support the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -1     | Force one-column output.                         |
   * | -a     | Show all entries, including "." entries.         |
   * | -c     | Show last status change time.                    |
   * | -C     | Force multi-column output.                       |
   * | -d     | List directories without recursion.              |
   * | -f     | Do not sort.                                     |
   * | -F     | Add special symbols on files and folders.        |
   * | -g     | Like -l, skip owner names (just show groups).    |
   * | -i     | Include the inode number.                        |
   * | -l     | Show a long listing format.                      |
   * | -L     | Show info on symbolic link target.               |
   * | -o     | Like -l, skip group names (just show owners).    |
   * | -q     | Replace non-printing characters with ?.          |
   * | -r     | Reverse the sort order.                          |
   * | -R     | Recursively list directories.                    |
   * | -s     | Include the size of each item.                   |
   * | -t     | Sort by modification time.                       |
   * | -u     | Sort by last access time.                        |
   *
   * Linux, macOS, and BSD support the following additional flags,
   * but not POSIX:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -h     | Simplify numbers with byte suffixes (e.g. MB).   |
   *
   * Linux and macOS support the following additional flags, but
   * not BSD:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -A     | Same as -a, but skip "." and "..".               |
   * | -H     | Follow symbolic links.                           |
   * | -k     | Show sizes in kilobytes, not blocks.             |
   * | -m     | Full width output with comma separated entries.  |
   * | -n     | Like -l, show numeric IDs.                       |
   * | -p     | Append '/' to directory names.                   |
   * | -S     | Sort by size.                                    |
   * | -x     | Order values side-to-side, instead of up-and-down.|
   *
   * Linux supports the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -b     | Show non-printing characters with C escape codes.|
   * | -B     | Ignore backup files ending with "~".             |
   * | -D     | Generate output for Emacs dired mode.            |
   * | -G     | With -l, skip group names.                       |
   * | -I     | Suppress entries matching a pattern.             |
   * | -N     | Print raw names, including control characters.   |
   * | -Q     | Enclose names in double-quotes.                  |
   * | -T     | Set tab stops.                                   |
   * | -U     | Do not sort.                                     |
   * | -v     | Use a natural sort for numbers in text.          |
   * | -w     | Set output width.                                |
   * | -X     | Sort by file name extension.                     |
   * | --all        | Same as -a.                                |
   * | --almost-all | Same as -A.                                |
   * | --author     | With -l, print the author of each file.    |
   * | --escape     | Same as -b.                                |
   * | --block-size | Show sizes in blocks.                      |
   * | --classify   | Same as -F.                                |
   * | --color      | Colorize output.                           |
   * | --dereference| Same as -L.                                |
   * | --dereference-command-line | Same as -H.                  |
   * | --dereference-command-line-symlink-to-dir| Follow symbolic links to dirs.|
   * | --directory  | Same as -d.                                |
   * | --dired      | Same as -D.                                |
   * | --file-type  | Same as -F, but don't add '*' on executables.|
   * | --format     | Customize formatting.                      |
   * | --full-time  | Same as -l, but use full ISO times.        |
   * | --help       | Show help on command.                      |
   * | --hide       | Hide entries that match a pattern.         |
   * | --hide-control-chars| Same as -q.                         |
   * | --human-readable | Same as -h.                            |
   * | --ignore     | Same as -I.                                |
   * | --ignore-backups | Same as -B.                            |
   * | --indicator-style| Same as -p, --file-type, or -F.        |
   * | --inode      | Same as -i.                                |
   * | --literal    | Same as -N.                                |
   * | --no-group   | Same as -G.                                |
   * | --quote-name | Same as -Q.                                |
   * | --quoting-style| Same as -Q but specify quoting.          |
   * | --numeric-uid-gid | Same as -n.                           |
   * | --recursive  | Same as -R.                                |
   * | --reverse    | Same as -r.                                |
   * | --show-control-characters| Show non-printing characters as-is.|
   * | --si         | Same as -h, but use powers of 1000 not 1024.|
   * | --size       | Same as -s.                                |
   * | --sort       | Select the sort key.                       |
   * | --tabsize    | Same as -T.                                |
   * | --time       | Select different time value.               |
   * | --time-style | Select time formatting.                    |
   * | --width      | Same as -w.                                |
   * | --version    | Show version information.                  |
   *
   * macOS supports the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -@     | With -l, show extended attributes.               |
   * | -b     | Show non-printing characters with octal escapes. |
   * | -B     | Same as -b, but use C escape codes too.          |
   * | -e     | With -l, show ACLs.                              |
   * | -G     | Colorize output.                                 |
   * | -O     | With -l, include file flags.                     |
   * | -P     | Show in on symbolic link, not target.            |
   * | -T     | With -l, show full detailed time.                |
   * | -U     | Sort by creation time.                           |
   * | -v     | Show non-printing characters as-is.              |
   * | -w     | Force raw printing of non-printable characters.  |
   * | -W     | Show whiteouts when scanning directories.        |
   *
   * For FolderShare, some of the above do not apply:
   *
   * - -b:  Linux/macOS "ls -b" shows names with non-printing characters using C
   *   escape codes. But FolderShare does not allow non-printing characters in
   *   names. A "-b" option has no meaning here.
   *
   * - -c:  Linux/macOS/BSD distinguish between when a file's content was
   *   last modified, and when its status (e.g. permissions, ownership) was
   *   last changed. "ls" defaults to showing the last content modified time.
   *   The "-c" option shifts to showing the last status change time.
   *   FolderShare only has one change time to show. A "-c" option has
   *   no meaning here.
   *
   * - -C:  Linux/macOS/BSD "ls" defaults to showing multi-column output on a
   *   terminal, but single-column output when redirected to a file. The "-C"
   *   option overrides this and forces multi-column output. This client,
   *   however, doesn't make a distinction between terminal and redirected
   *   output, so it always defaults to multi-column output. It only goes
   *   single column with the "-1" option. A "-C" option is redundant here.
   *
   * - -g:  Linux/macOS/BSD "ls" defaults to showing user names and group
   *   names in a "-l" output. The "-g" option suppresses the owner names
   *   and just shows group names. But FolderShare has no notion of account
   *   groups, so there are no group names to show.  A "-g" option has no
   *   meaning here.
   *
   * - -G:  Linux "ls -G" is the same as "ls -lo". It outputs a list that
   *   skips group names. But FolderShare has no notion of account groups,
   *   so there are no group names to show anyway. A "-G" option has no
   *   meaning here.
   *
   * - -H: Linux/macOS with "-H" will follow symbolic links. But FolderShare
   *   has no notion of symbolic links, so there is nothing to do with this.
   *   A "-H" option has no meaning here.
   *
   * - -k: Linux/macOS "ls" defaults to showing file/folder sizes in blocks.
   *   The "-k" option shifts to kilobytes, which is usually more useful.
   *   But FolderShare has no notion of blocks and always shows sizes in
   *   bytes. A "-k" option is redundant here.
   *
   * - -L:  Linux/macOS/BSD "ls" with "-l" shows symbolic links with the link's
   *   size and date. Add "-L" and it shows the symbolic link target's size
   *   and date instead. But FolderShare has no notion of symbolic links,
   *   so there is nothing to show. A "-L" option has no meaning here.
   *
   * - -N:  Linux "ls" with "-N" suppress showing non-printing characters as
   *   "?" characters and instead shows them as-is. Since FolderShare does not
   *   support non-printing characters, the "-N" option has nothing to do.
   *   The "-N" option is redundant here.
   *
   * - -o:  Linux/macOS/BSD "ls" defaults to showing user names and group
   *   names in a "-l" output. The "-o" option suppresses the group names
   *   and just shows owner names. But FolderShare has no notion of account
   *   groups, so there is nothing to suppress. A "-o" option is redundant.
   *
   * - -q: Linux/macOS/BSD "ls" defaults to showing non-printing characters as
   *   "?" characters when outputing to a terminal, and disables this when
   *   redirected to a file. The "-q" option forces redirected output to revert
   *   to terminal-style output. But this client does not make a distinction
   *   between terminal and redirected output. Further, FolderShare does not
   *   support non-printing characters in file names. A "-q" option has no
   *   meaning here.
   *
   * - -u: Linux/macOS/BSD "ls" defaults to sorting by name. The "-t" option
   *   sorts instead by modification date, and the "-u" option sorts by
   *   access time. But FolderShare does not maintain access times. A "-u"
   *   option has no meaning here.
   *
   * - -v: macOS "ls -v" is the same as Linux "ls -N" above and shows names
   *   as-is, even with non-printing characters. FolderShare does not support
   *   non-printing characters in names. A "-v" option has no meaning here.
   *
   * And so on. The above is enough to get the idea. Options that deal with
   * non-printing characters, creation times, symbolic links, or screen width
   * have no meaning here.
   */
  public static function ls(
    FolderShareConnect $server,
    array $options) {

    //
    // Parse options
    // -------------
    // Synonyms have already been mapped to primary flag names.
    $recurse = FALSE;
    $formatFlags = [];

    foreach ($options['flags'] as $flagName) {
      switch ($flagName) {
        case '-d':
          $recurse = FALSE;
          break;

        case '-R':
          $recurse = TRUE;
          break;

        case '-1':
        case '-f':
        case '-F':
        case '-h':
        case '-i':
        case '-l':
        case '-r':
        case '-s':
        case '-S':
        case '-t':
          $formatFlags[] = $flagName;
          break;
      }
    }

    //
    // Validate
    // --------
    // If there are no paths given, default to to the user's root list ('/').
    $remoteUris = $options['paths'];
    if (empty($remoteUris) === TRUE) {
      $remoteUris = ['/'];
    }
    else {
      $remoteUris = FolderShareConnect::normalizeUris($remoteUris, FALSE);
    }

    if ($recurse === TRUE &&
        $options['format'] !== 'linux') {
      self::printError(
        reset($options['command']),
        "Recursive lists must use a 'linux' output format.");
      return FALSE;
    }

    //
    // Execute
    // -------
    // Get a list of entities. Format the response using the flags, if any.
    try {
      while (empty($remoteUris) === FALSE) {
        $remoteUri = array_shift($remoteUris);

        if ($recurse === TRUE) {
          print("$remoteUri:\n");
        }

        // Get the list.
        switch ($remoteUri) {
          case '/':
          case '/personal':
            $response = $server->getRootItems('personal');
            break;

          case '/public':
            $response = $server->getRootItems('public');
            break;

          default:
            $response = $server->getDescendants($remoteUri);
            break;
        }

        // Print the response.
        if (empty($response) === FALSE) {
          switch ($options['format']) {
            case 'linux':
              print(FolderShareFormat::formatAsLinuxLs($response, $formatFlags));
              break;

            case 'text':
              print(FolderShareFormat::formatAsText($response));
              break;

            default:
            case 'full':
            case 'keyvalue':
              print_r($response);
          }
        }

        // Update the paths array by adding child paths to the front.
        if ($recurse === TRUE) {
          $recurseUris = [];
          if (is_array($response) === TRUE) {
            // Loop through the response and create a path for each folder item.
            foreach ($response as $r) {
              if (isset($r['path']) === TRUE && isset($r['kind']) === TRUE) {
                $k = $r['kind'];
                if ($k === 'folder') {
                  $recurseUris[] = $r['path'];
                }
              }
            }
          }

          // Add recursion paths to the front of the path list.
          // This causes us to do a depth-first traversal.
          $remoteUris = array_merge($recurseUris, $remoteUris);

          if (empty($remoteUris) === FALSE) {
            print("\n");
          }
        }
      }
    }
    catch (\Exception $e) {
      self::printException(reset($options['command']), $e);
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Prints file or folder status.
   *
   * This command emulates the Linux/macOS/BSD "stat" command.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -t     | Show a terse form of the information.            |
   * | --help | Show help on command.                            |
   * | --terse| Same as -t.                                      |
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @see ::chown()
   * @see ::ls()
   * @see ::update()
   * @see ::xattr()
   *
   * @internal
   * Linux "stat" supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -c     | Use a custom output format.                      |
   * | -f     | Show file system status instead of file status.  |
   * | -L     | Follow symbolic links.                           |
   * | -t     | Show a terse form of the information.            |
   * | --dereference | Same as -L.                               |
   * | --filesystem  | Same as -f.                               |
   * | --format      | Same as -c.                               |
   * | --help        | Show help on command.                     |
   * | --terse       | Same as -t.                               |
   * | --version     | Show version information.                 |
   *
   * BSD and macOS "stat" support the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -f     | Use a custom output format.                      |
   * | -F     | Add special symbols on files and folders.        |
   * | -l     | Same as "ls -lT".                                |
   * | -L     | Follow symbolic links.                           |
   * | -n     | Suppress newline after each item.                |
   * | -q     | Suppress error messages.                         |
   * | -r     | Show raw information from the 'stat' syscall.    |
   * | -s     | Show information in a shell compatible format.   |
   * | -t     | Display time stamps in a custom output format.   |
   * | -x     | Show a verbose output format similar to Linux.   |
   *
   * BSD supports the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -H     | Treat arguments as hex NFS file handles.         |
   *
   * There is no POSIX "stat" command.
   *
   * The set of flags for Linux and macOS/BSD are entirely disjoint -
   * there is nothing in common. Some flags, like -t, show up in both
   * versions of "stat", but with different meanings.
   */
  public static function stat(
    FolderShareConnect $server,
    array $options) {

    //
    // Parse flags
    // -----------
    // Synonyms have already been mapped to primary flag names.
    $terseMode = FALSE;

    foreach ($options['flags'] as $flagName) {
      switch ($flagName) {
        case '-t':
          $terseMode = TRUE;
          break;
      }
    }

    //
    // Validate
    // --------
    // If there are no remote URIs given, default to the user's
    // root list at '/'.
    $remoteUris = $options['paths'];
    if (empty($remoteUris) === TRUE) {
      $remoteUris = ['/'];
    }
    else {
      $remoteUris = FolderShareConnect::normalizeUris($remoteUris);
    }

    //
    // Execute
    // -------
    // Get entity information. Do not recurse into folders.
    try {
      foreach ($remoteUris as $remoteUri) {
        $response = $server->getFileOrFolder($remoteUri, FALSE);

        if (empty($response) === FALSE) {
          switch ($options['format']) {
            case 'linux':
              print(FolderShareFormat::formatAsLinuxStat($response, $terseMode));
              break;

            case 'text':
            case 'keyvalue':
              print(FolderShareFormat::formatAsText($response));
              break;

            case 'full':
            default:
              print_r($response);
              break;
          }
        }
      }
    }
    catch (\Exception $e) {
      self::printException(reset($options['command']), $e);
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Waits until a file or folder becomes enabled.
   *
   * This command does not emulate any particular Linux/macOS/BSD
   * command, though it is very roughly modeled after the "sync"
   * command.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | --help | Show help on command.                            |
   * | -v     | Be verbose.                                      |
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @see ::waitUntilEnabled()
   */
  public static function wait(
    FolderShareConnect $server,
    array $options) {

    //
    // Parse flags
    // -----------
    // Synonyms have already been mapped to primary flag names.
    $verbose = FALSE;

    foreach ($options['flags'] as $flagName) {
      switch ($flagName) {
        case '-v':
          $verbose = TRUE;
          break;
      }
    }

    //
    // Validate
    // --------
    // If there are no paths given, default to '/'.
    $remoteUris = $options['paths'];
    if (empty($remoteUris) === TRUE) {
      $remoteUris = ['/'];
    }
    else {
      $remoteUris = FolderShareConnect::normalizeUris($remoteUris);
    }

    //
    // Execute
    // -------
    // Repeatedly get each entity and check if they are all enabled.
    // If any one of them is not, wait and repeat.
    self::waitUntilEnabled(
      $server,
      reset($options['command']),
      $remoteUris,
      $verbose);
    return TRUE;
  }

  /**
   * Prints extended attributes values.
   *
   * This command does not emulate any particular Linux/macOS/BSD
   * command, though it is named after the macOS "xattr" command.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | --help | Show help on command.                            |
   * | --ATTR | The name of an attribute.                        |
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @see ::ls()
   * @see ::stat()
   * @see ::update()
   */
  public static function xattr(
    FolderShareConnect $server,
    array $options) {

    //
    // Parse flags
    // -----------
    // Synonyms have already been mapped to primary flag names.
    $attrs = [];

    foreach ($options['flags'] as $flagName) {
      switch ($flagName) {
        case '--disabled':
          $attrs['systemdisabled'] = 'bool';
          break;

        case '--hidden':
          $attrs['systemhidden'] = 'bool';
          break;

        case '--kind':
          $attrs['kind'] = 'string';
          break;

        case '--mime':
          $attrs['mime'] = 'string';
          break;

        case '--size':
          $attrs['size'] = 'int';
          break;
      }
    }

    if (empty($attrs) === TRUE) {
      // Default to all available attributes.
      $attrs = [
        'systemdisabled' => 'bool',
        'systemhidden'   => 'bool',
        'kind'           => 'string',
        'mime'           => 'string',
        'size'           => 'int',
      ];
    }

    $nAttrs = count($attrs);

    //
    // Validate
    // --------
    // If there are no paths given, default to '/'.
    $remoteUris = $options['paths'];
    if (empty($remoteUris) === TRUE) {
      $remoteUris = ['/'];
    }
    else {
      $remoteUris = FolderShareConnect::normalizeUris($remoteUris);
    }

    //
    // Execute
    // -------
    // Get entity information.
    try {
      foreach ($remoteUris as $remoteUri) {
        $response = $server->getFileOrFolder($remoteUri, FALSE);

        foreach ($attrs as $attr => $cast) {
          if (isset($response[$attr]) === TRUE &&
              isset($response[$attr][0]) === TRUE &&
              isset($response[$attr][0]['value']) === TRUE) {
            $value = $response[$attr][0]['value'];
            if ($nAttrs > 1) {
              print("$attr: ");
            }

            switch ($cast) {
              case 'bool':
                if (empty($value) === TRUE || $value === FALSE) {
                  print("false\n");
                }
                else {
                  print("true\n");
                }
                break;

              case 'string':
                print((string) $value . "\n");
                break;

              case 'int':
                print((int) $value . "\n");
                break;
            }
          }
        }
      }
    }
    catch (\Exception $e) {
      self::printException(reset($options['command']), $e);
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Updates file and folder field values.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -v     | Show the name of each item modified.             |
   * | --help | Show help on command.                            |
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @see ::chown()
   * @see ::ls()
   * @see ::stat()
   * @see ::xattr()
   *
   * @internal
   * There is no standard Linux/macOS/BSD command for changing file metadata.
   *
   * The Linux 'getfattr' and 'setfattr' commands get and set arbitrary file
   * metadata:
   * - getfattr -d files...
   *   Print all metadata.
   *
   * - getfattr -n fieldname files...
   *   Print the metadata for the selected field.
   *
   * - setfattr -n fieldname -v fieldvalue files...
   *   Write the metadata for the selected field.
   *
   * - setfattr -x fieldname files...
   *   Delete the metadata for the selected field.
   *
   * The Linux commands support the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -e     | Set text encoding.                               |
   * | -h     | Apply to link, not target of link.               |
   * | -m     | Only show fields that match a pattern.           |
   * | -R     | Recurse into directories.                        |
   * | -L     | Follow symbolic links.                           |
   * | -P     | Do not follow symbolic links.                    |
   * | --absolute-name | Don't remove leading '/'.               |
   * | --dump    | Same as -d.                                   |
   * | --encoding| Same as -e.                                   |
   * | --help    | Show help on command.                         |
   * | --logical | Same as -L.                                   |
   * | --match   | Same as -m.                                   |
   * | --no-dereference| Same as -h.                             |
   * | --only-values | Only print out field values.              |
   * | --physical| Same as -P.                                   |
   * | --recursive| Same as -r.                                  |
   * | --version | Show version information.                     |
   *
   * The macOS 'xattr' command gets, sets, and deletes arbitrary file metadata:
   * - xattr files...
   *   Print all metadata.
   *
   * - xattr -p fieldname files...
   *   Print the metadata for the selected field.
   *
   * - xattr -w fieldname fieldvalue files...
   *   Write the metadata for the selected field.
   *
   * - xattr -d fieldname files...
   *   Delete the metadata for the selected field.
   *
   * - xattr -c files...
   *   Clear all metadata.
   *
   * The macOS commands support the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -l     | Long form that shows field names and values.     |
   * | -r     | Recurse into directories.                        |
   * | -s     | Apply to link, not target of link.               |
   * | -v     | Show the name of each item modified.             |
   * | -x     | Print output in hex.                             |
   *
   * The BSD 'getextattr', 'lsextattr', 'rmextattr', and 'setextattr' commands
   * get, list, delete, and set arbitrary file metadata:
   * - lsextattr namespace files...
   *   Print all metadata. 'namespace' is either 'user' or 'system'.
   *
   * - getextattr namespace fieldname files...
   *   Print the metadata for the selected field.
   *
   * - setextattr namespace fieldname fieldvalue files...
   *   Write the metadata for the selected field.
   *
   * - rmextattr namespace fieldname files...
   *   Delete the metadata for the selected field.
   *
   * The BSD commands support the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -f     | Don't prompt for comfirmation and ignore errors. |
   * | -h     | Apply to link, not target of link.               |
   * | -i     | Read attributes from stdin.                      |
   * | -n     | NULL terminate output data.                      |
   * | -q     | Be quiet, don't print path name or errors.       |
   * | -s     | Convert non-printing characters to escapes.      |
   * | -x     | Print output in hex.                             |
   *
   * This FolderShare "update" supports a structure similar to the Linux,
   * macOS, and BSD commands, but not identical to any of them.
   *
   * - FolderShare does not have arbitrary metadata - the fields have
   *   to be predefined by the FolderShare module, by third-party
   *   modules, or by the site administrator. So Linux/macOS/BSD-style
   *   operations to create or delete metadata fields are not meaningful.
   *
   * - The 'stat' command already outputs all entity fields. All we need
   *   here is a command to update them.
   *
   * - The Linux, macOS, and BSD commands all have a form that gives
   *   the name of a field, the value, and a list of files. We use
   *   that same form.
   */
  public static function update(
    FolderShareConnect $server,
    array $options) {

    //
    // Parse options
    // -------------
    // Synonyms have already been mapped to primary flag names.
    $verbose = FALSE;

    foreach ($options['flags'] as $flagName) {
      switch ($flagName) {
        case '-v':
          $verbose = TRUE;
          break;
      }
    }

    //
    // Validate
    // --------
    // The command must have at least:
    // - A field name.
    // - A field value.
    // - A path.
    //
    // All further items are paths for additional entities.
    $remoteUris = FolderShareConnect::normalizeUris($options['paths']);
    switch (count($remoteUris)) {
      case 0:
        self::printError(
          reset($options['command']),
          "Missing field name, value, and remote file path.");
        return FALSE;

      case 1:
        self::printError(
          reset($options['command']),
          "Missing field value and remote file path.");
        return FALSE;

      case 2:
        self::printError(
          reset($options['command']),
          "Missing remote file path.");
        return FALSE;

      case 3:
        $fieldName = array_shift($remoteUris);
        $fieldValue = array_shift($remoteUris);
        break;

      default:
        self::printError(
          reset($options['command']),
          "Too many arguments.");
        return FALSE;
    }

    //
    // Execute
    // -------
    // Change a series of items.
    try {
      foreach ($remoteUris as $remoteUri) {
        if ($verbose === TRUE) {
          print("$remoteUri\n");
        }

        // There is normally no response.
        $response = $server->update($remoteUri, $fieldName, $fieldValue);

        if (empty($response) === FALSE) {
          print_r($response);
        }
      }
    }
    catch (\Exception $e) {
      self::printException(reset($options['command']), $e);
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Changes the owner of files and folders.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -R     | Recursively change descendents.                  |
   * | -v     | Show the name of each item modified.             |
   * | --help | Show help on command.                            |
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @see ::ls()
   * @see ::stat()
   * @see ::xattr()
   */
  public static function chown(
    FolderShareConnect $server,
    array $options) {

    //
    // Parse options
    // -------------
    // Synonyms have already been mapped to primary flag names.
    $verbose = FALSE;
    $recurse = FALSE;

    foreach ($options['flags'] as $flagName) {
      switch ($flagName) {
        case '-R':
          $recurse = TRUE;
          break;

        case '-v':
          $verbose = TRUE;
          break;
      }
    }

    //
    // Validate
    // --------
    // The command must have at least:
    // - A user name or ID.
    // - A path.
    //
    // All further items are paths for additional entities.
    $userName = '';
    switch (count($options['paths'])) {
      case 0:
        self::printError(
          reset($options['command']),
          "Missing user name and remote file path.");
        return FALSE;

      case 1:
        self::printError(
          reset($options['command']),
          "Missing remote file path.");
        return FALSE;

      default:
        break;
    }

    $userName = array_shift($options['paths']);
    $remoteUris = FolderShareConnect::normalizeUris($options['paths']);

    //
    // Execute
    // -------
    // Change a series of items.
    try {
      foreach ($remoteUris as $remoteUri) {
        if ($verbose === TRUE) {
          print("$remoteUri\n");
        }

        // There is normally no response.
        $response = $server->changeOwner($userName, $remoteUri, $recurse);

        if (empty($response) === FALSE) {
          print_r($response);
        }
      }
    }
    catch (\Exception $e) {
      self::printException(reset($options['command']), $e);
      return FALSE;
    }

    return TRUE;
  }

  /*--------------------------------------------------------------------
   *
   * File and folder delete commands.
   *
   *--------------------------------------------------------------------*/

  /**
   * Deletes files and folders.
   *
   * This command emulates the Linux/macOS/BSD "rm" command.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -d     | Remove directories as well as files.             |
   * | -f     | Don't prompt for comfirmation and ignore errors. |
   * | -R     | Recursively delete directory trees (implies -d). |
   * | -r     | Same as -R.                                      |
   * | -v     | Print the name of each item as it is deleted.    |
   * | --help | Show help on command.                            |
   *
   * By default, this method does not remove folders. Use "-d"
   * to remove a file or folder.
   *
   * By default, this method does not recurse and will not remove
   * folders that are not empty. Use "-r" to recurse.
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @internal
   * POSIX, Linux, macOS, and BSD "rm" support the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -f     | Don't prompt for comfirmation and ignore errors. |
   * | -i     | Interactively confirm each item before deletion. |
   * | -R     | Recursively delete directory trees (implies -d). |
   * | -r     | Same as -R.                                      |
   *
   * Linux, macOS, and BSD "rm" support the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -v     | Print the name of each item as it is deleted.    |
   *
   * macOS and BSD "rm" support the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -d     | Remove directories as well as files.             |
   * | -P     | Overwrite regular files before deleting them.    |
   * | -W     | Undelete white-out deleted files.                |
   *
   * BSD "rm" supports the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -I     | Interactively confirm deletes > 3 files or dirs. |
   * | -x     | Do not cross mount points during recursive delete.|
   *
   * All Linux versions of "rm" support the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -I     | Interactively confirm deletes > 3 files or dirs. |
   * | --force            | Same as -f.                          |
   * | --help             | Show help on command.                |
   * | --interactive=WHEN | Same as -i and -I.                   |
   * | --no-preserve-root | Don't prevent '/' delete.            |
   * | --one-file-system  | Same as BSD -x.                      |
   * | --preserve-root    | Do prevent '/' delete.               |
   * | --recursive        | Same as -R.                          |
   * | --verbose          | Same as -v.                          |
   * | --version          | Show version information.            |
   *
   * RedHat Linux "rm" supports the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -d     | Remove directories as well as files.             |
   * | --directory        | Same as -d.                          |
   *
   * CentOS Linux "rm" supports the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -d     | Remove directories as well as files.             |
   * | --dir  | Same as -d.                                      |
   */
  public static function rm(
    FolderShareConnect $server,
    array $options) {

    //
    // Parse flags
    // -----------
    // Synonyms have already been mapped to primary flag names.
    $verbose      = FALSE;
    $recurse      = FALSE;
    $fileOrFolder = FALSE;
    $ignoreErrors = FALSE;

    foreach ($options['flags'] as $flagName) {
      switch ($flagName) {
        case '-d':
          $fileOrFolder = TRUE;
          break;

        case '-f':
          $ignoreErrors = TRUE;
          break;

        case '-r':
          $recurse = TRUE;
          $fileOrFolder = TRUE;
          break;

        case '-v':
          $verbose = TRUE;
          break;
      }
    }

    //
    // Validate
    // --------
    // There must be at least one path.
    if (empty($options['paths']) === TRUE) {
      self::printError(
        reset($options['command']),
        "Missing remote file path.");
      return FALSE;
    }

    //
    // Execute
    // -------
    // Remove each of the indicated entities.
    $remoteUris = FolderShareConnect::normalizeUris($options['paths']);
    foreach ($remoteUris as $remoteUri) {
      if ($verbose === TRUE) {
        print("$remoteUri\n");
      }

      try {
        // There is normally no response.
        if ($fileOrFolder === FALSE) {
          $response = $server->deleteFile($remoteUri);
        }
        else {
          $response = $server->deleteFileOrFolder($remoteUri, $recurse);
        }

        if (empty($response) === FALSE) {
          print_r($response);
        }
      }
      catch (\Exception $e) {
        if ($ignoreErrors === TRUE) {
          // The -f flag was given, which blocks does-not-exist error
          // messages and setting the exit code to failure on messages.
          continue;
        }

        self::printException(reset($options['command']), $e);
        return FALSE;
      }
    }

    return TRUE;
  }

  /**
   * Deletes empty folders.
   *
   * This command emulates the Linux/macOS/BSD "rmdir" command.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -v     | Show the name of each item deleted.              |
   * | --help | Show help on command.                            |
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @internal
   * POSIX, Linux, macOS, and BSD "rmdir" support the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -p     | Remove every directory on the path.              |
   *
   * Linux supports the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -v     | Show the name of each item deleted.              |
   * | --help    | Show help on command.                         |
   * | --ignore-fail-on-non-empty | Ignore non-empty errors.     |
   * | --parents | Same as -p.                                   |
   * | --verbose | Same as -v.                                   |
   * | --version | Show version information.                     |
   *
   * BSD supports the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -v     | Show the name of each item deleted.              |
   */
  public static function rmdir(
    FolderShareConnect $server,
    array $options) {

    //
    // Parse flags
    // -----------
    // Synonyms have already been mapped to primary flag names.
    $verbose = FALSE;

    foreach ($options['flags'] as $flagName) {
      switch ($flagName) {
        case '-v':
          $verbose = TRUE;
          break;
      }
    }

    //
    // Validate
    // --------
    // There must be at least one path.
    if (empty($options['paths']) === TRUE) {
      self::printError(
        reset($options['command']),
        "Missing remote file path.");
      return FALSE;
    }

    //
    // Execute
    // -------
    // Remove each of the indicated entities.
    $remoteUris = FolderShareConnect::normalizeUris($options['paths']);
    try {
      foreach ($remoteUris as $remoteUri) {
        if ($verbose === TRUE) {
          print("$remoteUri\n");
        }

        // There is normally no response.
        $response = $server->deleteFolder($remoteUri);

        if (empty($response) === FALSE) {
          print_r($response);
        }
      }
    }
    catch (\Exception $e) {
      self::printException(reset($options['command']), $e);
      return FALSE;
    }

    return TRUE;
  }

  /*--------------------------------------------------------------------
   *
   * Folder create commands.
   *
   *--------------------------------------------------------------------*/

  /**
   * Creates a new root folder or subfolder.
   *
   * This command emulates the Linux/macOS/BSD "mkdir" command.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -p     | Create intermediate folders as required.         |
   * | -v     | Show the name of each item created.              |
   * | --help | Show help on command.                            |
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @internal
   * POSIX, Linux, macOS, and BSD "mkdir" support the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -m     | Set the permissions mode for new directories.    |
   * | -p     | Create parent directories too, if needed.        |
   * | -v     | Show the name of each item created.              |
   *
   * Linux has the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | --help    | Show help on command.                         |
   * | --mode    | Same as -m.                                   |
   * | --parents | Same as -p.                                   |
   * | --verbose | Same as -v.                                   |
   * | --version | Show version information.                     |
   */
  public static function mkdir(
    FolderShareConnect $server,
    array $options) {

    //
    // Parse options
    // -------------
    // Synonyms have already been mapped to primary flag names.
    $verbose = FALSE;
    $createAncestors = FALSE;

    foreach ($options['flags'] as $flagName) {
      switch ($flagName) {
        case '-p':
          $createAncestors = TRUE;
          break;

        case '-v':
          $verbose = TRUE;
          break;
      }
    }

    //
    // Validate
    // --------
    // There must be at least one path.
    if (empty($options['paths']) === TRUE) {
      self::printError(
        reset($options['command']),
        "Missing remote file path.");
      return FALSE;
    }

    $remoteUris = FolderShareConnect::normalizeUris($options['paths']);

    //
    // Execute
    // -------
    // Create a series of folders. None of these should already exist.
    // If one does, report an error.
    //
    // For each path, build a list of folders to create for that path.
    // Without "-p", this will just be the last folder on the path.
    // But with "-p", this may include any of the intermediate folders
    // that do not exist.
    foreach ($remoteUris as $remoteUri) {
      $foldersToCreate = [];

      // We have these cases:
      // - The path is only a /. Report an error.
      // - The path does not start with a /. Report an error.
      // - The only / is at the start. Create a root folder.
      // - The last slash is anywhere in the middle. Create a subfolder.
      if ($remoteUri === '/') {
        self::printError(
          reset($options['command']),
          "The remote \"/\" folder already exists.");
        return FALSE;
      }

      $firstCharacter = mb_substr($remoteUri, 0, 1);
      if ($firstCharacter !== '/') {
        self::printError(
          reset($options['command']),
          "Paths must start with '/'.");
        return FALSE;
      }

      $indexOfLastSlash = mb_strrpos($remoteUri, '/');

      if ($indexOfLastSlash === 0) {
        // The path has the form "/ROOT" without further slashes.
        // Verify that it doesn't exist.
        try {
          $response = $server->getFileOrFolder($remoteUri, FALSE);

          self::printError(
            reset($options['command']),
            "A file or folder \"$remoteUri\" already exists.");
          return FALSE;
        }
        catch (\Exception $e) {
          // The root folder doesn't exist. Add it to the creation list.
          $foldersToCreate[] = [
            'parent' => '/',
            'name'   => mb_substr($remoteUri, 1),
          ];
        }
      }
      elseif ($createAncestors === FALSE) {
        // The path has the form "/ROOT/CHILD...". Do not create ancestors,
        // so verify that they all exist and are folders.
        $parent = mb_substr($remoteUri, 0, $indexOfLastSlash);
        $name = mb_substr($remoteUri, ($indexOfLastSlash + 1));

        try {
          $response = $server->getFileOrFolder($parent, FALSE);

          if ($response['kind'] !== 'folder') {
            // Parent is not a folder!
            self::printError(
              reset($options['command']),
              "\"$parent\" is not a folder.");
            return FALSE;
          }
        }
        catch (\Exception $e) {
          $commandName = reset($options['command']);
          print("$commandName: " . strip_tags($e->getMessage()) . "\n");
          return FALSE;
        }

        try {
          $response = $server->getFileOrFolder($remoteUri, FALSE);

          self::printError(
            reset($options['command']),
            "A remote file or folder \"$remoteUri\" already exists.");
          return FALSE;
        }
        catch (\Exception $e) {
          // The parent exists, but the child doesn't. Add the child to
          // the creation list.
          $foldersToCreate[] = [
            'parent' => $parent,
            'name'   => $name,
          ];
        }
      }
      else {
        // The path has the form "/ROOT/CHILD...". Do create ancestors
        // as needed. Loop over all of the ancestors and add them to list
        // to create if they don't already exist. If they do exist, make
        // sure they are folders.
        $names = mb_split('/', $remoteUri);
        array_shift($names);
        $parent = '/';
        $ancestorsExist = TRUE;

        // Loop through the names on the path, starting with the root.
        $n = count($names);
        for ($i = 0; $i < $n; ++$i) {
          $name = $names[$i];
          $path = ($parent === '/') ? ('/' . $name) : ($parent . '/' . $name);
          $errorIfExists = ($i === ($n - 1));

          if ($ancestorsExist === TRUE) {
            // The ancestors so far exist, but the new child might not.
            // Test first.
            try {
              $response = $server->getFileOrFolder($path, FALSE);

              if ($errorIfExists === TRUE) {
                // The last item on the path already exists, report an error.
                self::printError(
                  reset($options['command']),
                  "A file or folder \"$path\" already exists.");
                return FALSE;
              }

              if ($response['kind'] !== 'folder') {
                // Ancestor is not a folder!
                self::printError(
                  reset($options['command']),
                  "\"$parent\" is not a folder.");
                return FALSE;
              }

              // Otherwise it is OK that the ancestor exists. Don't add it
              // to the list of folders to create and keep going to check
              // the next name on the path.
            }
            catch (\Exception $e) {
              // The child doesn't exist. Add it to the list and note that we
              // found an ancestor that doesn't exist.  All further children
              // won't exist either, so there is no further need to test them.
              $foldersToCreate[] = [
                'parent' => $parent,
                'name'   => $name,
              ];
              $ancestorsExist = FALSE;
            }
          }
          else {
            // The ancestors don't exist yet, so the child can't exist
            // either. Just add it to the list.
            $foldersToCreate[] = [
              'parent' => $parent,
              'name'   => $name,
            ];
          }

          $parent = $path;
        }
      }

      // Now create each of the folders. Usually there will be just one,
      // but if "-p" was given, there may be multiple folders to create.
      foreach ($foldersToCreate as $folderInfo) {
        $name = $folderInfo['name'];
        $parent = $folderInfo['parent'];
        if ($parent === '/') {
          $path = '/' . $name;
        }
        else {
          $path = $parent . '/' . $name;
        }

        if ($verbose === TRUE) {
          print("Create folder \"$path\"\n");
        }

        try {
          // There is normally no response from newRootFolder() or newFolder().
          if ($parent === '/') {
            $response = $server->newRootFolder($name);
          }
          else {
            $response = $server->newFolder($parent, $name);
          }

          if (empty($response) === FALSE) {
            print_r($response);
          }
        }
        catch (\Exception $e) {
          // The folder could not be created. Perhaps it already exists.
          // Or perhaps the user does not have sufficient permissions.
          self::printException(reset($options['command']), $e);
          return FALSE;
        }
      }
    }

    return TRUE;
  }

  /*--------------------------------------------------------------------
   *
   * File or folder copy and move commands.
   *
   *--------------------------------------------------------------------*/

  /**
   * Moves a file or folder to a new location.
   *
   * This command emulates the Linux/macOS/BSD "mv" command.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -n     | Do not overwrite an existing file.               |
   * | -v     | Show the name of each item moved.                |
   * | --help | Show help on command.                            |
   * | --wait | Wait for move to complete.                       |
   *
   * This command always recursively copies folders.
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @internal
   * POSIX, Linux, macOS, and BSD "mv" support the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -f     | Don't prompt for comfirmation and ignore errors. |
   * | -i     | Interactively confirm each item before moving.   |
   *
   * Linux, macOS, and BSD have the following additional flags, but
   * not POSIX:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -n     | Do not overwrite an existing file.               |
   * | -v     | Show the name of each item moved.                |
   *
   * BSD has the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -h     | Do not follow target symbolic links.             |
   *
   * Linux has the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -S     | Override the usual backup suffix.                |
   * | -t     | Move everything into the target directory.       |
   * | -T     | Treat the target as a file, not a directory.     |
   * | -u     | Only move if the source is newer.                |
   * | --force       | Same as -f.                               |
   * | --help    | Show help on command.                         |
   * | --interactive | Same as -i.                               |
   * | --no-clobber  | Same as -n.                               |
   * | --no-target-directory | Same as -T.                       |
   * | --suffix      | Same as -S.                               |
   * | --target-directory | Same as -t.                          |
   * | --update  | Same as -u.                                   |
   * | --verbose | Same as -v.                                   |
   * | --version | Show version information.                     |
   */
  public static function mv(
    FolderShareConnect $server,
    array $options) {

    //
    // Parse options
    // -------------
    // Synonyms have already been mapped to primary flag names.
    $verbose = FALSE;
    $overwrite = TRUE;
    $wait = FALSE;

    foreach ($options['flags'] as $flagName) {
      switch ($flagName) {
        case '-n':
          $overwrite = FALSE;
          break;

        case '-v':
          $verbose = TRUE;
          break;

        case '--wait':
          $wait = TRUE;
          break;
      }
    }

    //
    // Validate
    // --------
    // The command has one of two forms:
    // - mv source destination
    // - mv source1 source2 source3... destination
    //
    // The first form is a degenerate case of the second that supports
    // optional renaming of the source. In that case the destination does
    // not refer to an entity that exists, yet.
    //
    // However, only the server knows whether a destination exists. So
    // we need to break this down into a loop of source-to-destination
    // operations and watch for failure.
    //
    // Confirm there are at least two paths given.
    $remoteUris = FolderShareConnect::normalizeUris($options['paths']);
    switch (count($remoteUris)) {
      case 0:
        self::printError(
          reset($options['command']),
          "Missing remote file paths.");
        return FALSE;

      case 1:
        self::printError(
          reset($options['command']),
          "Missing remote destination file path.");
        return FALSE;

      default:
        $destinationUri = array_pop($remoteUris);
        break;
    }

    //
    // Execute
    // -------
    // Move a series of items to the same destination. Abort on the first error.
    foreach ($remoteUris as $remoteUri) {
      if ($verbose === TRUE) {
        print("$remoteUri\n");
      }

      try {
        // There is normally no response.
        $response = $server->move($remoteUri, $destinationUri, $overwrite);
        if (empty($response) === FALSE) {
          print_r($response);
        }
      }
      catch (\Exception $e) {
        self::printException(reset($options['command']), $e);
        return FALSE;
      }
    }

    if ($wait === TRUE) {
      self::waitUntilEnabled(
        $server,
        reset($options['command']),
        [$destinationUri],
        $verbose);
    }

    return TRUE;
  }

  /**
   * Copies a file or folder to a new location.
   *
   * This command emulates the Linux/macOS/BSD "cp" command.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -n     | Do not overwrite existing files.                 |
   * | -v     | Show the name of each item copied.               |
   * | --help | Show help on command.                            |
   * | --wait | Wait for copy to complete.                       |
   *
   * This command always recursively copies folders.
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @internal
   * POSIX, Linux, macOS, and BSD "cp" support the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -f     | Don't prompt for comfirmation and ignore errors. |
   * | -H     | With -R, follow command line symbolic links.     |
   * | -i     | Interactively confirm each item before copying.  |
   * | -L     | With -R, follow recursive symbolic links.        |
   * | -P     | With -R, don't follow symbolic links (default).  |
   * | -p     | Preserve dates, permissions, and owner on copy.  |
   * | -R     | Recursively copy trees.                          |
   *
   * Linux, macOS, and BSD have the following additional flags, but
   * not POSIX:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -a     | Same as -pPR.                                    |
   * | -n     | Do not overwrite existing files.                 |
   * | -v     | Show the name of each item copied.               |
   *
   * BSD has the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -l     | Hard link files instead of copying.              |
   * | -x     | Do not cross mount points during recursive copy. |
   *
   * macOS has the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -c     | Copy files useing clonefile.                     |
   * | -X     | Do not copy extended attributes.                 |
   *
   * Linux has the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -b     | Make a backup of each destination file.          |
   * | -d     | Same as -P --preserve=links.                     |
   * | -l     | Hard link files instead of copying.              |
   * | -n     | Do not overwrite existing files.                 |
   * | -r     | Same as -R.                                      |
   * | -s     | Symbolic link files instead of copying.          |
   * | -S     | Override backup file name suffix.                |
   * | -t     | Specify destination (target) directory.          |
   * | -T     | Treat destination as a file, not a directory.    |
   * | -u     | Copy only if the source file is newer.           |
   * | -x     | Do not cross mount points during recursive copy. |
   * | --archive| Same as -a.                                    |
   * | --backup | Same as -b.                                    |
   * | --copy-contents | Copy special files when recursive.      |
   * | --dereference   | Same as -L.                             |
   * | --force         | Same as -f.                             |
   * | --help | Show help on command.                            |
   * | --interactive   | Same as -i.                             |
   * | --link          | Same as -l.                             |
   * | --no-clobber    | Same as -n.                             |
   * | --no-dereference| Same as -P.                             |
   * | --no-preserve   | Don't preserve attributes on copy.      |
   * | --no-target-directory| Same as -T.                        |
   * | --parents       | Use full source file name under dir.    |
   * | --one-file-system| Same as -x.                            |
   * | --preserve      | Preserve specific attributes on copy.   |
   * | --recursive     | Same as -R.                             |
   * | --reflink       | Control clones.                         |
   * | --remove-directories| Remove existing destinations first. |
   * | --sparse        | Control sparse files.                   |
   * | --strip-trailing-slashes| Remove last slashes in source.  |
   * | --symbolic-link | Same as -s.                             |
   * | --suffix        | Same as -S.                             |
   * | --target-directory| Same as -t.                           |
   * | --update        | Same as -u.                             |
   * | --verbose       | Same as -v.                             |
   * | --version       | Show version information.               |
   */
  public static function cp(
    FolderShareConnect $server,
    array $options) {

    //
    // Parse options
    // -------------
    // Synonyms have already been mapped to primary flag names.
    $verbose = FALSE;
    $overwrite = TRUE;
    $wait = FALSE;

    foreach ($options['flags'] as $flagName) {
      switch ($flagName) {
        case '-n':
          $overwrite = FALSE;
          break;

        case '-v':
          $verbose = TRUE;
          break;

        case '--wait':
          $wait = TRUE;
          break;
      }
    }

    //
    // Validate
    // --------
    // The command has one of two forms:
    // - cp source destination
    // - cp source1 source2 source3... destination
    //
    // The first form is a degenerate case of the second that supports
    // optional renaming of the copy. In that case the destination does
    // not refer to an entity that exists, yet.
    //
    // However, only the server knows whether a destination exists. So
    // we need to break this down into a loop of source-to-destination
    // operations and watch for failure.
    //
    // Confirm there are at least two paths given.
    $remoteUris = FolderShareConnect::normalizeUris($options['paths']);
    switch (count($remoteUris)) {
      case 0:
        self::printError(
          reset($options['command']),
          "Missing remote file paths.");
        return FALSE;

      case 1:
        self::printError(
          reset($options['command']),
          "Missing remote destination file path.");
        return FALSE;

      default:
        $destinationUri = array_pop($remoteUris);
        break;
    }

    //
    // Execute
    // -------
    // Copy a series of items to the same destination. Abort on the first error.
    foreach ($remoteUris as $remoteUri) {
      if ($verbose === TRUE) {
        print("$remoteUri\n");
      }

      try {
        // There is normally no response.
        $response = $server->copy($remoteUri, $destinationUri, $overwrite);
        if (empty($response) === FALSE) {
          print_r($response);
        }
      }
      catch (\Exception $e) {
        self::printException(reset($options['command']), $e);
        return FALSE;
      }
    }

    if ($wait === TRUE) {
      self::waitUntilEnabled(
        $server,
        reset($options['command']),
        [$destinationUri],
        $verbose);
    }

    return TRUE;
  }

  /*--------------------------------------------------------------------
   *
   * File or folder download and upload.
   *
   *--------------------------------------------------------------------*/

  /**
   * Downloads a file and folder.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -n     | Do not overwrite existing files.                 |
   * | -p     | Preserve dates on download.                      |
   * | -u     | Copy only if the source file is newer.           |
   * | -v     | Show the name of each item downloaded.           |
   * | --help | Show help on command.                            |
   * | --sync | Same as -pu.                                     |
   *
   * The default behavior is to overwrite local files with downloaded files.
   * The -n option blocks overwriting and instead adds a unique numeric sequence
   * number onto the base name of the downloaded file (e.g. "file-2.txt",
   * "file-3.txt", ...).
   *
   * The default is for newly downloaded files and folders to have the
   * current date and time as their timestamps. The -p option uses the
   * remote item's date and time instead.
   *
   * The default is to download all files, overwriting if allowed to.
   * The -u option causes downloads to occur only if the remote file is
   * newer or if there is no local file downloaded yet with the same name.
   * If used together with -n, which prevents overwriting, -nu only
   * downloads missing files.
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @see ::cp()
   * @see ::upload()
   *
   * @internal
   * There is no standard Linux/macOS/BSD command for downloading a file.
   * If we use 'cp' as a rough model, POSIX, Linux, macOS, and BSD "cp"
   * support the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -f     | Don't prompt for comfirmation and ignore errors. |
   * | -H     | With -R, follow command line symbolic links.     |
   * | -i     | Interactively confirm each item before copying.  |
   * | -L     | With -R, follow recursive symbolic links.        |
   * | -P     | With -R, don't follow symbolic links (default).  |
   * | -p     | Preserve dates, permissions, and owner on copy.  |
   * | -R     | Recursively copy trees.                          |
   *
   * Linux, macOS, and BSD have the following additional flags, but
   * not POSIX:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -a     | Same as -pPR.                                    |
   * | -n     | Do not overwrite existing files.                 |
   * | -v     | Show the name of each item copied.               |
   *
   * BSD has the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -l     | Hard link files instead of copying.              |
   * | -x     | Do not cross mount points during recursive copy. |
   *
   * macOS has the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -c     | Copy files useing clonefile.                     |
   * | -X     | Do not copy extended attributes.                 |
   *
   * Linux has the following additional flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -b     | Make a backup of each destination file.          |
   * | -d     | Same as -P --preserve=links.                     |
   * | -l     | Hard link files instead of copying.              |
   * | -n     | Do not overwrite existing files.                 |
   * | -r     | Same as -R.                                      |
   * | -s     | Symbolic link files instead of copying.          |
   * | -S     | Override backup file name suffix.                |
   * | -t     | Specify destination (target) directory.          |
   * | -T     | Treat destination as a file, not a directory.    |
   * | -u     | Copy only if the source file is newer.           |
   * | -x     | Do not cross mount points during recursive copy. |
   * | --archive| Same as -a.                                    |
   * | --backup | Same as -b.                                    |
   * | --copy-contents | Copy special files when recursive.      |
   * | --dereference   | Same as -L.                             |
   * | --force         | Same as -f.                             |
   * | --help | Show help on command.                            |
   * | --interactive   | Same as -i.                             |
   * | --link          | Same as -l.                             |
   * | --no-clobber    | Same as -n.                             |
   * | --no-dereference| Same as -P.                             |
   * | --no-preserve   | Don't preserve attributes on copy.      |
   * | --no-target-directory| Same as -T.                        |
   * | --parents       | Use full source file name under dir.    |
   * | --one-file-system| Same as -x.                            |
   * | --preserve      | Preserve specific attributes on copy.   |
   * | --recursive     | Same as -R.                             |
   * | --reflink       | Control clones.                         |
   * | --remove-directories| Remove existing destinations first. |
   * | --sparse        | Control sparse files.                   |
   * | --strip-trailing-slashes| Remove last slashes in source.  |
   * | --symbolic-link | Same as -s.                             |
   * | --suffix        | Same as -S.                             |
   * | --target-directory| Same as -t.                           |
   * | --update        | Same as -u.                             |
   * | --verbose       | Same as -v.                             |
   * | --version       | Show version information.               |
   */
  public static function download(
    FolderShareConnect $server,
    array $options) {

    //
    // Parse options
    // -------------
    // Synonyms have already been mapped to primary flag names.
    $verbose = FALSE;
    $overwrite = TRUE;
    $preserve = FALSE;
    $update = FALSE;

    foreach ($options['flags'] as $flagName) {
      switch ($flagName) {
        case '-n':
          $overwrite = FALSE;
          break;

        case '-p':
          $preserve = TRUE;
          break;

        case '--sync':
          $update = TRUE;
          $preserve = TRUE;
          break;

        case '-u':
          $update = TRUE;
          break;

        case '-v':
          $verbose = TRUE;
          break;
      }
    }

    $downloadFlags = 0;
    if ($overwrite === TRUE) {
      $downloadFlags |= FolderShareConnect::DOWNLOAD_OVERWRITE;
    }
    if ($preserve === TRUE) {
      $downloadFlags |= FolderShareConnect::DOWNLOAD_PRESERVE;
    }
    if ($update === TRUE) {
      $downloadFlags |= FolderShareConnect::DOWNLOAD_UPDATE;
    }

    //
    // Validate
    // --------
    // The command needs two paths:
    // - URI to remote file or folder.
    // - Path to local file or folder.
    //
    // If the local path is not given, it defaults to the current directory.
    switch (count($options['paths'])) {
      case 0:
        self::printError(
          reset($options['command']),
          "Missing remote file path.");
        return FALSE;

      case 1:
        // One path: Remote. Use current directory as local path.
        $remoteUri = $options['paths'][0];
        $localPath = getcwd();
        if ($localPath === FALSE) {
          // On some UNIX variants, getcwd() returns FALSE if one of the
          // current directory's ancestors is not readable/searchable. This
          // is sufficient to cause failure of the download.
          self::printError(
            reset($options['command']),
            "The current directory is not readable so it cannot be used to receive a download.");
          return FALSE;
        }
        break;

      case 2:
        // Two paths: Remote & Local.
        $remoteUri = $options['paths'][0];
        $localPath = $options['paths'][1];
        break;

      default:
        self::printError(
          reset($options['command']),
          "Too many file paths.");
        return FALSE;
    }

    $remoteUri = FolderShareConnect::normalizeUri($remoteUri);

    //
    // Execute
    // -------
    // Download the file.
    try {
      $priorProgress = $server->isProgress();

      if ($verbose === TRUE) {
        print("Download \"$remoteUri\" to \"$localPath\"\n");
        $server->setProgress(TRUE);
      }

      $server->download($remoteUri, $localPath, $downloadFlags);

      $server->setProgress($priorProgress);
    }
    catch (\Exception $e) {
      if ($verbose === TRUE) {
        print("\n");
      }

      self::printException(reset($options['command']), $e);
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Uploads a local file or folder.
   *
   * This method supports the following flags:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -n     | Do not overwrite existing files.                 |
   * | -p     | Preserve dates on download.                      |
   * | -u     | Copy only if the source file is newer.           |
   * | -v     | Show the name of each item uploaded.             |
   * | --help | Show help on command.                            |
   * | --sync | Same as -pu.                                     |
   *
   * The default behavior is to overwrite local files with downloaded files.
   * The -n option blocks overwriting and instead adds a unique numeric sequence
   * number onto the base name of the downloaded file (e.g. "file-2.txt",
   * "file-3.txt", ...).
   *
   * The default is for newly downloaded files and folders to have the
   * current date and time as their timestamps. The -p option uses the
   * remote item's date and time instead.
   *
   * The default is to download all files, overwriting if allowed to.
   * The -u option causes downloads to occur only if the remote file is
   * newer or if there is no local file downloaded yet with the same name.
   * If used together with -n, which prevents overwriting, -nu only
   * downloads missing files.
   *
   * @param \FolderShare\FolderShareConnect $server
   *   The server connection.
   * @param array $options
   *   The command-line options array.
   *
   * @return bool
   *   Returns TRUE on success, and FALSE on failure. On failure an error
   *   message has already been output.
   *
   * @see ::cp()
   * @see ::download()
   *
   * @internal
   * There is no standard Linux/macOS/BSD command for uploading a file.
   * If we use 'cp' as a rough model. See runDownload() comments for more.
   */
  public static function upload(
    FolderShareConnect $server,
    array $options) {

    //
    // Parse options
    // -------------
    // Synonyms have already been mapped to primary flag names.
    $verbose = FALSE;
    $overwrite = TRUE;
    $preserve = FALSE;
    $update = FALSE;

    foreach ($options['flags'] as $flagName) {
      switch ($flagName) {
        case '-n':
          $overwrite = FALSE;
          break;

        case '-p':
          $preserve = TRUE;
          break;

        case '--sync':
          $update = TRUE;
          $preserve = TRUE;
          break;

        case '-u':
          $update = TRUE;
          break;

        case '-v':
          $verbose = TRUE;
          break;
      }
    }

    $uploadFlags = 0;
    if ($overwrite === TRUE) {
      $uploadFlags |= FolderShareConnect::UPLOAD_OVERWRITE;
    }
    if ($preserve === TRUE) {
      $uploadFlags |= FolderShareConnect::UPLOAD_PRESERVE;
    }
    if ($update === TRUE) {
      $uploadFlags |= FolderShareConnect::UPLOAD_UPDATE;
    }

    //
    // Validate
    // --------
    // The command needs two paths:
    // - Path to local file or folder.
    // - URI to remote file or folder.
    switch (count($options['paths'])) {
      case 0:
        self::printError(
          reset($options['command']),
          "Missing local and remote file paths.");
        return FALSE;

      case 1:
        self::printError(
          reset($options['command']),
          "Missing remote file path.");
        break;

      case 2:
        // Two paths: Remote & Local.
        $localPath = $options['paths'][0];
        $remoteUri = $options['paths'][1];
        break;

      default:
        self::printError(
          reset($options['command']),
          "Too many file paths.");
        return FALSE;
    }

    $remoteUri = FolderShareConnect::normalizeUri($remoteUri, FALSE);

    //
    // Execute
    // -------
    // Upload the file or folder.
    try {
      // There is normally no response.
      $priorProgress = $server->isProgress();
      $server->setProgress($verbose);

      if ($verbose === TRUE) {
        print("Upload \"$localPath\" to \"$remoteUri\"...\n");
      }

      $server->upload($localPath, $remoteUri, $uploadFlags);

      $server->setProgress($priorProgress);
    }
    catch (\Exception $e) {
      if ($verbose === TRUE) {
        print("\n");
      }

      self::printException(reset($options['command']), $e);
      return FALSE;
    }

    return TRUE;
  }

}
