<?php

namespace FolderShare;

/**
 * Provides static functions that assist in specialized content formatting.
 *
 * Intended as a companion class to FolderShareConnect, this class provides
 * optional formatting utility functions to present returned content in
 * special ways.
 */
class FolderShareFormat {

  const MAX_SCREEN_WIDTH = 80;

  /*--------------------------------------------------------------------
   *
   * Key-value formatting.
   *
   *--------------------------------------------------------------------*/

  /**
   * Formats a nested key-value array as text.
   *
   * The keys and values of the array are returned as multi-line text in
   * two columns holding the key and value. Rows are optionally indented.
   *
   * Nested arrays are handled by recursing.
   *
   * @param array $response
   *   The key-value response to convert to text.
   * @param string $indent
   *   (optional, default = '  ') The indent for each key-value line in
   *   the result.
   *
   * @return string
   *   The string representation.
   */
  public static function formatAsText(
    array $response,
    string $indent = '') {

    //
    // Validate
    // --------
    // It is not an error to try and format an empty response array.
    // It could simply mean that there is no response.
    if (empty($response) === TRUE) {
      return '';
    }

    //
    // Prepare
    // -------
    // Make a first pass through all keys to find the longest key.
    // This will become the column width below.
    $maxLength = 0;
    foreach (array_keys($response) as $key) {
      $length = mb_strlen($key);
      if ($length > $maxLength) {
        $maxLength = $length;
      }
    }

    // Add one for a ':'.
    ++$maxLength;

    //
    // Generate output
    // ---------------
    // Make a second pass to format everything. Use the maximum key length
    // to set a uniform field size for outputing key names.
    $format = $indent . '%-' . $maxLength . "s %s\n";
    $text = '';
    foreach ($response as $key => $value) {
      // If the value is a scalar, convert it to a string and output.
      if (is_scalar($value) === TRUE) {
        $text .= sprintf($format, $key . ':', $value);
        continue;
      }

      if (empty($value) === TRUE) {
        $text .= sprintf($format, $key . ':', '');
        continue;
      }

      // Otherwise the value must be an array. Sweep through the array to
      // see if all of its values are scalars and its keys are numeric.
      // This is the case for simple lists.
      $simpleList = TRUE;
      foreach ($value as $k => $v) {
        if (is_int($k) === FALSE || is_scalar($v) === FALSE) {
          $simpleList = FALSE;
          break;
        }
      }

      // If the array is a simple list, format the array as a comma-separated
      // list.
      if ($simpleList === TRUE) {
        $text .= sprintf($format, $key . ':', implode(', ', $value));
        continue;
      }

      // Otherwise recurse.
      $indentMore = $indent . '  ';
      $text .= sprintf($format, $key . ':', '');
      $text .= self::formatAsText($value, $indentMore);
    }

    return $text;
  }

  /*--------------------------------------------------------------------
   *
   * Linux-style formatting.
   *
   *--------------------------------------------------------------------*/

  /**
   * Formats a single entity similar to the output of a Linux 'stat' command.
   *
   * The Linux 'stat' command prints information about a file, directory,
   * or device. It has two primary formats:
   *
   * - (default) A full format shows multiple lines of text with key-value
   *   pairs giving the file name, size, permissions, etc.
   *
   * - A terse format shows a single line of text that condenses the most
   *   important values, including the file name, size, permissions, etc.
   *
   * This method takes 'stat' key-value data from FolderShare and formats
   * it in a full-like or terse-like output that includes essential values
   * for a FolderShare entity. It does not support format control arguments.
   *
   * Linux, BSD, macOS, and others all have 'stat' commands, but their
   * formats differ. This method outputs in a format similar to Linux.
   *
   * The Linux 'stat' output is not particularly user-friendly. Linux users
   * are more likely to be familiar with 'ls' output, which provides similar
   * information.
   *
   * <B>Full mode (default)</B><BR>
   * The full multi-line output of Linux 'stat' includes:
   * - File name.
   * - Size.
   * - Number of blocks.
   * - IO block size.
   * - File type (e.g. "regular file" vs. "device").
   * - Device.
   * - Inode number.
   * - Number of hard links.
   * - Permissions mode as hex and text.
   * - User as ID and account name.
   * - Group as ID and group name.
   * - Time of last access as text.
   * - Time of last modification as text.
   * - Time of last status change as text.
   *
   * Some of these values have FolderShare equivalents, while others do not.
   * This method drops these values:
   * - Number of blocks.
   * - IO block size.
   * - Number of hard links.
   * - Group as ID and group name.
   * - Time of last access as text.
   *
   * And replaces these values:
   * - File type replaced with MIME type.
   * - Device replaced with host name.
   * - Inode number replaced with entity ID.
   * - Time of last status change replaced with creation date.
   *
   * This method also adds a few fields:
   * - Full path.
   * - Parent entity ID.
   * - Root entity ID.
   *
   * This method also slightly adjusts the Linux 'stat' names to be more
   * meaningful here:
   * - "Name" is used instead of the misleading "File", which Linux uses
   *   even for things that are not files.
   * - "Mime type" is used instead of the Linux "FileType".
   *
   * The output of this method looks like:
   * <pre>
   *   Path: "PATH"
   *   Name: "FILENAME"
   *   Size: SIZE         Mime type: TYPE
   *   Host: HOSTNAME     ID: ID  ParentID: ID  RootID: ID
   * Access: SHARING      Uid: (UID/ACCOUNT)
   * Modify: DATE
   * Create: DATE
   * </pre>
   *
   * <B>Terse mode</B><BR>
   * If the $terseMode argument is TRUE, a terse mode is returned instead
   * of the full version above.
   *
   * The terse output of Linux 'stat' is an alias for the format string:
   * - %n %s %b %f %u %g %D %i %h %t %T %X %Y %Z %W %o
   * where:
   * - %n = file name.
   * - %s = size.
   * - %b = number of blocks.
   * - %f = raw permission mode in hex.
   * - %u = user ID.
   * - %g = group ID.
   * - %D = device number in hex.
   * - %i = inode number.
   * - %h = number of hard links.
   * - %t = major device type in hex.
   * - %T = minor device type in hex.
   * - %X = time of last access as a timestamp.
   * - %Y = time of last modification as a timestamp.
   * - %Z = time of last status change as a timestamp.
   * - %o = optimal I/O transfer size hint.
   *
   * Some of these values have FolderShare equivalents, while others do not.
   * Rather than reduce the line to just the FolderShare values, this method
   * retains the columns of the Linux terse mode, but shows a '-' if the
   * value is not relevant.
   *
   * This method shows the following values:
   * - %n = file name.
   * - %s = size.
   * - %b = -.
   * - %f = sharing status.
   * - %u = user ID.
   * - %g = -.
   * - %D = host name.
   * - %i = entity ID.
   * - %h = -.
   * - %t = -.
   * - %T = -.
   * - %X = -.
   * - %Y = time of last modification as a timestamp.
   * - %Z = time of creation as a timestamp.
   * - %o = -.
   *
   * @param array $response
   *   The key-value array response when querying a file or folder.
   * @param bool $terseMode
   *   (optional, default = FALSE) When TRUE, a terse one-line form of
   *   the information is created. Otherwise a multi-line form is created.
   *
   * @return string
   *   Returns a text representation of the response.
   *
   * @throws \InvalidArgumentException
   *   Throws an exception if the $response array is not for an entity or
   *   list of entities.
   */
  public static function formatAsLinuxStat(
    array $response,
    bool $terseMode = FALSE) {
    //
    // Validate
    // --------
    // Make sure the given server response is key-value data for
    // a single entity or a list of entities.
    if (empty($response) === TRUE) {
      throw new \InvalidArgumentException(
        "Programmer error: Empty response array when formatting as 'stat'.");
    }

    if (isset($response['path']) === FALSE) {
      // The array is not for a single entity. Is it for a list of entities?
      foreach ($response as $r) {
        if (isset($r['path']) === FALSE) {
          throw new \InvalidArgumentException(
            "Programmer error: Response array is not in key-value form when formatting as 'stat'.");
        }
      }

      // Recurse to build the output for a list of entities.
      $text = '';
      foreach ($response as $r) {
        $text .= self::formatAsLinuxStat($r, $terseMode);
      }

      return $text;
    }

    //
    // Generate output
    // ---------------
    // Extract and format relevant values.
    //
    // The 'size' may not be present if a entity does not yet have a size
    // computed.
    //
    // The 'id' may not be present if the entity is a temporary false entity
    // for "/" or some other virtual entity.
    //
    // All other values should be present.
    //
    // Entity fields.
    $size     = (isset($response['size']) === TRUE) ?
      (string) $response['size'] : '0';
    $mimeType = (isset($response['mime']) === TRUE) ?
      $response['mime'] : '-';
    $id       = (isset($response['id']) === TRUE) ?
      (string) $response['id'] : '-';
    $parentId = (isset($response['parentid']) === TRUE) ?
      (string) $response['parentid'] : '-';
    $rootId   = (isset($response['rootid']) === TRUE) ?
      (string) $response['rootid'] : '-';

    // Virtual fields.
    $path        = $response['path'];
    $permission  = $response['sharing-status'];
    $uid         = (isset($response['uid']) === TRUE ?
      $response['uid'] :
      (isset($response['user-id']) === TRUE ?
      $response['user-uid'] : '-'));
    $accountName = (isset($response['uid-account-name']) === TRUE ?
      $response['uid-account-name'] :
      (isset($response['user-account-name']) === TRUE ?
      $response['user-account-name'] : '-'));

    // Create formatted dates. The server provides two forms of each date:
    //
    // - The raw timestamp stored on the server. This is expected to be a
    //   UNIX timestamp with an epoch of Jan 1 1970. However, depending upon
    //   this may be risky because different OSes use different epochs, and
    //   the current host running this client may use a different epoch.
    //
    // - An ISO 8601 format time that includes the year, month, day, hour,
    //   minutes, seconds, and timezone offset in a well-defined parsable
    //   format that is independent of the timestamp epoch.
    //
    // The ISO 8601 time is the one to use. However, it is certainly not
    // in the Linux stat date format. So convert the ISO 8601 date to a
    // local timestamp, and the format that.
    //
    // Formatting codes for date() are:
    // - D = 3-letter day of week.
    // - M = 3-letter month abbreviation.
    // - j = date of week without leading zeroes.
    // - H = hour on 24-hour clock without leading zeroes.
    // - i = minutes with leading zeroes.
    // - s = seconds with leading zeroes.
    // - Y = 4-digit year.
    //
    // Get the current timezone. It should have been set in PHP's .ini
    // file as "date.timezone", or it could have been set explicitly during
    // this script run.
    $tz = date_default_timezone_get();
    $modified = @strtotime($response['changed-date']);
    $created = @strtotime($response['created-date']);

    if ($modified === FALSE) {
      // An older version of the REST protocol had a weak formatted date
      // that may not be parsable. Just use it as-is.
      $modifiedDate = $response['changed-date'];
    }
    elseif ($tz === 'UTC') {
      // UTC is the default and rarely what the user expects. Add it to
      // the formatted date.
      $modifiedDate = date('D M j H:i:s Y e', $modified);
    }
    else {
      // Otherwise use the default timezone for the system.
      $modifiedDate = date('D M j H:i:s Y', $modified);
    }

    if ($created === FALSE) {
      // An older version of the REST protocol had a weak formatted date
      // that may not be parsable. Just use it as-is.
      $createdDate = $response['created-date'];
    }
    elseif ($tz === 'UTC') {
      // UTC is the default and rarely what the user expects. Add it to
      // the formatted date.
      $createdDate = date('D M j H:i:s Y e', $created);
    }
    else {
      // Otherwise use the default timezone for the system.
      $createdDate = date('D M j H:i:s Y', $created);
    }

    // Create simplified file type per Linux.
    // - Root and subfolders = "Directory".
    // - Everything else = "File".
    if ($mimeType === 'folder/directory') {
      $fileType = "Directory ($mimeType)";
    }
    else {
      $fileType = "Regular File ($mimeType)";
    }

    // Site info.
    $host = $response['host'];

    // Get the file name from the path. We cannot use PHP's 'basename'
    // since it uses the current OS's conventions ("locale") rather than
    // our generic conventions. We also need to be multi-byte safe.
    if ($path === '/') {
      $filename = $path;
    }
    else {
      $lastSlashPosition = mb_strrpos($path, '/');
      if ($lastSlashPosition === FALSE) {
        $filename = $path;
      }

      $filename = mb_substr($path, ($lastSlashPosition + 1));
    }

    // Format the output.
    $text = '';
    if ($terseMode === TRUE) {
      $text .= sprintf(
        "%s %s - %s %d - - %d - - - - %d %d -\n",
        $filename,
        $size,
        $permission,
        $uid,
        $id,
        $modified,
        $created);
    }
    else {
      // Formatting here approximately mimics Linux et al.
      // - %6s = the first-in-line label, which Linux shows as <= 6 chars.
      //
      // - %-13s = the indent to the 2nd column, which is also used in Linux.
      //   This length is dictated by the length of human-readable dates in
      //   the Modify/Create lines.
      //
      // File: "FILE"
      // - Mimics Linux. Like Linux, it always uses the "File" label, even
      //   when the item is a folder. Linux shows whatever value is given
      //   to the command (e.g. "./stuff" or "/a/b/c/stuff"). Since we
      //   do not support relative paths, the full path is always shown.
      $text .= sprintf(
        "%6s: \"%s\"\n",
        'File',
        $path);

      // Host: NAME
      // - FolderShare-specific. We cannot add other items to the line
      //   because host/domain names can be long and it would mess up
      //   further line formatting.
      $text .= sprintf(
        "%6s: %-13s\n",
        'Host',
        $host);

      // Size: SIZE  FileType: TYPE (MIME)
      // - Mimics Linux. Spacing to the FileType differs - increased so that
      //   items line up with the next few multi-value lines.
      // - Linux shows the FileType as either "Regular File" or "Directory"
      //   and doesn't include a Mime type. We add the Mime type in parens.
      $text .= sprintf(
        "%6s: %-13s %-9s %s\n",
        'Size',
        $size,
        'FileType:',
        $fileType);

      // ID: ID  ParentID: ID  RootID:  ID
      // - FolderShare-specific. Linux instead shows the device major/minor
      //   numbers, Inode number, and the number of links. We show entity IDs.
      $text .= sprintf(
        "%6s: %-13s %-9s %-9s %s %-9s\n",
        'ID',
        $id,
        'ParentID:',
        $parentId,
        'RootID:',
        $rootId);

      // Mode: STATUS  UID: (UID/LOGIN)
      // - Mimics Linux. Instead of the RWX permissions groups of Linux,
      //   we show the generic sharing status.
      // - Linux formats the UID and account name as (%5d/%8s) because the
      //   original account name length limit was 8 characters, and the
      //   original account ID limit was for a 16-bit number. Linux still
      //   uses this formatting, but overflows this presentation as needed.
      $text .= sprintf(
        "%6s: %-13s %s: (%5d/%8s)\n",
        'Mode',
        $permission,
        'Uid',
        $uid,
        $accountName);

      // Modify: DATE
      // - Mimics Linux.
      $text .= sprintf(
        "%6s: %s\n",
        'Modify',
        $modifiedDate);

      // Create: DATE
      // - FolderShare-specific. Linux et al do not store the creation date.
      //   Instead it has a "Change" time that records the last time inode
      //   data was set (owner, group, link count, mode).
      $text .= sprintf(
        "%6s: %s\n",
        'Create',
        $createdDate);
    }

    return $text;
  }

  /**
   * Formats an entity list similar to the output of a Linux 'ls' command.
   *
   * The Linux 'ls' command prints a list of files and folders. In its
   * default mode, file and folder names are shown in multiple columns
   * sized to fit on the screen. Several common options modify this output
   * to list one file per line and include more information. Some of the
   * most useful options are:
   * - '-l' = show a long form that adds the file's permissions, number of
   *   hard links, owner name, group name, size (in blocks), modified date,
   *   and file name.
   *
   * - '-s' = add the file's size in bytes.
   *
   * - '-i' = add the Inode number.
   *
   * - '-o' = omit the group name with paired with '-l'..
   *
   * There are many more options to Linux 'ls', but they are more obscure
   * and not relevant here.
   *
   * For this method, several options are supported to create a few common
   * output modes:
   *
   * - Short mode. File and folder names are output in mulitple columns
   *   sized to fit within an 80-character wide window. Output is sorted
   *   by the file name. If '-s' or '-i' options are given, the size and
   *   entity ID are included before each name.
   *
   * - Long mode. File and folder names are output with one per line,
   *   including attributes in a Linux style. Lines are sorted by the
   *   file name. The '-s' option is ignored because sizes are always
   *   shown in bytes. If '-i' is given, the entity ID is included at
   *   the start of each line. The group name is always skipped, as if
   *   '-o' were given. The number of hard links is always skipped
   *   since there is no such thing in FolderShare.
   *
   * The following values are available to show in short or long modes:
   * - Entity ID in place of Inode number.
   * - Sharing status in place of Linux permissions.
   * - User account name.
   * - Size (always in bytes).
   * - Modification date.
   * - Name.
   *
   * The following formatting flags are supported:
   * | Flag   | Meaning                                          |
   * | ------ | ------------------------------------------------ |
   * | -1     | Force one-column output.                         |
   * | -f     | Do not sort.                                     |
   * | -F     | Add special symbols on files and folders.        |
   * | -h     | Simplify numbers with byte suffixes (e.g. MB).   |
   * | -i     | Include the inode number.                        |
   * | -l     | Show a long listing format.                      |
   * | -r     | Reverse the sort order.                          |
   * | -s     | Include the size of each item.                   |
   * | -S     | Sort by size.                                    |
   * | -t     | Sort by modification time.                       |
   *
   * @param array $response
   *   The key-value array response when querying a file or folder.
   * @param array $flags
   *   (optional, default = []) Linux 'ls' style command-line flags.
   *
   * @return string
   *   Returns a text representation of the response.
   *
   * @throws \InvalidArgumentException
   *   Throws an exception if the $response array is not for an entity or
   *   list of entities.
   */
  public static function formatAsLinuxLs(array $response, array $flags = []) {
    //
    // Validate
    // --------
    // Make sure the given server response is key-value data for
    // a single entity or a list of entities.
    if (empty($response) === TRUE) {
      // It is not an error to try and format an empty response array.
      // It could simply mean that a folder listing found nothing to list.
      return '';
    }

    if (isset($response['path']) === TRUE) {
      // Response is for a single entity. Make it an array of one entry
      // and continue.
      $response = [$response];
    }
    else {
      // The array is not for a single entity. Is it for a list of entities?
      foreach ($response as $r) {
        if (isset($r['path']) === FALSE) {
          throw new \InvalidArgumentException(
            "Programmer error: Response array is not in key-value form when formatting for 'ls'.");
        }
      }
    }

    //
    // Check options
    // -------------
    // Look for options we support.
    $longMode = FALSE;
    $showSize = FALSE;
    $showId = FALSE;
    $showSymbol = FALSE;
    $sortBy = 'name';
    $sortOrder = "increasing";
    $sortReverse = FALSE;
    $multiColumn = TRUE;
    $simplifySizes = FALSE;

    foreach ($flags as $flag) {
      switch ($flag) {
        case '-1':
          $multiColumn = FALSE;
          break;

        case '-f':
          $sortOrder = "none";
          break;

        case '-F':
          $showSymbol = TRUE;
          break;

        case '-h':
          $simplifySizes = TRUE;
          break;

        case '-i':
          $showId = TRUE;
          break;

        case '-l':
          $longMode = TRUE;
          $multiColumn = FALSE;
          break;

        case '-r':
          $sortReverse = TRUE;
          break;

        case '-s':
          $showSize = TRUE;
          break;

        case '-S':
          $sortBy = 'size';
          break;

        case '-t':
          $sortBy = 'changed';
          $sortOrder = "decreasing";
          break;
      }
    }

    if ($sortReverse === TRUE && $sortOrder !== "none") {
      if ($sortOrder === "increasing") {
        $sortOrder = "decreasing";
      }
      else {
        $sortOrder = "increasing";
      }
    }

    $n = count($response);

    //
    // Sort
    // ----
    // Sort the response list. The sort is case sensitive, so upper case
    // names bubble to the top of the list.
    if ($sortOrder !== "none") {
      switch ($sortBy) {
        default:
        case 'name':
          if ($sortOrder === "increasing") {
            usort(
              $response,
              function ($a, $b) {
                $aname = basename($a['path']);
                $bname = basename($b['path']);
                return ($aname < $bname) ? (-1) : 1;
              });
          }
          else {
            usort(
              $response,
              function ($a, $b) {
                $aname = basename($a['path']);
                $bname = basename($b['path']);
                return ($aname > $bname) ? (-1) : 1;
              });
          }
          break;

        case 'size':
          if ($sortOrder === "increasing") {
            usort(
              $response,
              function ($a, $b) {
                if (isset($a['size']) === TRUE) {
                  $asize = (float) $a['size'];
                }
                else {
                  $asize = 0;
                }

                if (isset($b['size']) === TRUE) {
                  $bsize = (float) $b['size'];
                }
                else {
                  $bsize = 0;
                }

                return ($asize < $bsize) ? (-1) : 1;
              });
          }
          else {
            usort(
              $response,
              function ($a, $b) {
                if (isset($a['size']) === TRUE) {
                  $asize = (float) $a['size'];
                }
                else {
                  $asize = 0;
                }

                if (isset($b['size']) === TRUE) {
                  $bsize = (float) $b['size'];
                }
                else {
                  $bsize = 0;
                }

                return ($asize > $bsize) ? (-1) : 1;
              });
          }
          break;

        case 'changed':
          if ($sortOrder === "increasing") {
            usort(
              $response,
              function ($a, $b) {
                $atime = $a['changed'];
                $btime = $b['changed'];
                return ($atime < $btime) ? (-1) : 1;
              });
          }
          else {
            usort(
              $response,
              function ($a, $b) {
                $atime = $a['changed'];
                $btime = $b['changed'];
                return ($atime > $btime) ? (-1) : 1;
              });
          }
          break;
      }
    }

    //
    // Generate output
    // ---------------
    // Extract and format relevant values.
    //
    // The 'size' may not be present if a folder does not yet have a size
    // computed. All other values are always present.
    if ($longMode === FALSE) {
      // Output a listing that primarily includes names. Use multiple
      // columns if flagged to do so.
      $filenames = [];
      $maxLength = 0;

      // Loop through the response and find the longest file name
      // and, if needed, the symbol suffix based on the kind.
      foreach ($response as $r) {
        // Get the item name.
        $name = basename($r['path']);

        // Get a symbol suffix based on the kind, if any.
        if ($showSymbol === TRUE) {
          $kind = $r['kind'];
          if ($kind === 'folder') {
            $name .= '/';
          }
        }

        // Save the name for the list below.
        $filenames[] = $name;

        // Keep track of the longest name.
        $length = mb_strlen($name);
        if ($length > $maxLength) {
          $maxLength = $length;
        }
      }

      // If the size and entity ID are included in the output, increase
      // the maximum length to account for the space they require.
      if ($showSize === TRUE) {
        $maxLength += (8 + 1);
      }

      if ($showId === TRUE) {
        $maxLength += (8 + 1);
      }

      if ($multiColumn === TRUE) {
        // Use multi-column output.
        //
        // Add one for a space between columns.
        ++$maxLength;

        // Compute the number of columns of names in the output.
        // Allow a maximum of 80 characters to a line.
        //
        // (80 is an historical line length that dates back to Hollerith punch
        // card widths and FORTRAN, but it remains a common magic number of
        // terminal window widths).
        $nColumns = ceil(self::MAX_SCREEN_WIDTH / $maxLength);
        $nRows = (int) ($n / $nColumns);
        if (($nRows * $nColumns) !== $n) {
          ++$nRows;
        }
      }
      else {
        // Use single-column output.
        $nColumns = 1;
        $nRows = $n;
      }

      // Generate output.
      $text = '';
      for ($i = 0; $i < $nRows; ++$i) {
        $rowText = '';
        for ($j = 0; $j < $nColumns; ++$j) {
          // Compute the index of the next response.
          $index = (($i * $nColumns) + $j);
          if ($index >= $n) {
            continue;
          }

          // Get values.
          $filename = $filenames[$index];
          $size = isset($response[$index]['size']) === TRUE ?
            (string) $response[$index]['size'] : 0;
          $id = $response[$index]['id'];

          // Format the item.
          $itemText = '';
          if ($showId === TRUE) {
            $itemText .= sprintf("%8d ", $id);
          }

          if ($showSize === TRUE) {
            if ($simplifySizes === TRUE) {
              $itemText .= sprintf("%8s ", self::formatBytes($size));
            }
            else {
              $itemText .= sprintf("%8s ", $size);
            }
          }

          $itemText .= sprintf("%-" . $maxLength . "s", $filename);
          if ($j === 0) {
            $rowText .= $itemText;
          }
          else {
            $rowText .= ' ' . $itemText;
          }
        }

        $text .= $rowText . "\n";
      }

      return $text;
    }

    // Do a first pass to determine the maximum width of the permission
    // (sharing status) column and the account name column. The sharing
    // status can be as wide as 15 characters, but it is often less.
    // And the account name column can be as wide as 60 characters (Drupal
    // core maximum), but it is almost always less.
    $maxPermissionWidth = 1;
    $maxAccountNameWidth = 1;

    foreach ($response as $item) {
      // Get values.
      $permission = $item['sharing-status'];
      $accountName = (isset($item['uid-account-name']) === TRUE ?
        $item['uid-account-name'] :
        (isset($item['user-account-name']) === TRUE ?
        $item['user-account-name'] : '-'));

      if (empty($accountName) === TRUE) {
        // The anonymous account has no account name. Revert to its
        // display name, which better exist.
        $accountName = (isset($item['uid-display-name']) === TRUE ?
          $item['uid-display-name'] :
          (isset($item['user-display-name']) === TRUE ?
          $item['user-display-name'] : '-'));
        if (empty($accountName) === TRUE) {
          $accountName = "unknown";
        }
      }

      $len = mb_strlen($permission);
      if ($len > $maxPermissionWidth) {
        $maxPermissionWidth = $len;
      }

      $len = mb_strlen($accountName);
      if ($len > $maxAccountNameWidth) {
        $maxAccountNameWidth = $len;
      }
    }

    // Generate output.
    $text = '';
    foreach ($response as $item) {
      // Get values.
      $filename     = basename($item['path']);
      $size         = isset($item['size']) === TRUE ? (string) $item['size'] : '-';
      $id           = $item['id'];
      $permission   = $item['sharing-status'];
      $accountName = (isset($item['uid-account-name']) === TRUE ?
        $item['uid-account-name'] :
        (isset($item['user-account-name']) === TRUE ?
        $item['user-account-name'] : '-'));

      $modified = strtotime($item['changed-date']);
      $tz = date_default_timezone_get();
      if ($tz === 'UTC') {
        // UTC is the default timezone when nothing is set. This is rarely
        // what the user expects. Add the timezone name to the end.
        $modifiedDate = date('M D H:i e', $modified);
      }
      else {
        // Otherwise format it using the current timezone.
        $modifiedDate = date('M D H:i', $modified);
      }

      if (empty($accountName) === TRUE) {
        // The anonymous account has no account name. Revert to its
        // display name, which better exist.
        $accountName = (isset($item['uid-display-name']) === TRUE ?
          $item['uid-display-name'] :
          (isset($item['user-display-name']) === TRUE ?
          $item['user-display-name'] : '-'));
        if (empty($accountName) === TRUE) {
          $accountName = "unknown";
        }
      }

      // Get a symbol suffix based on the kind, if any.
      if ($showSymbol === TRUE) {
        $kind = $item['kind'];
        if ($kind === 'folder') {
          $filename .= '/';
        }
      }

      // Format the row.
      $idText = '';
      if ($showId === TRUE) {
        $idText = sprintf("%8d ", $id);
      }

      if ($simplifySizes === TRUE) {
        $size = self::formatBytes($size);
      }

      // The maximum "permission" (really sharing status) is 15 characters.
      //
      // The maximum account name is *60* characters (see Drupal core's
      // UserInterface::USERNAME_MAX_LENGTH). But we can't really leave
      // a 60 character field or the list will look odd. Instead, use
      // 15 characters.
      $text .= sprintf(
        "%s%" . $maxPermissionWidth . "s %" . $maxAccountNameWidth . "s %8s %s %s\n",
        $idText,
        $permission,
        $accountName,
        $size,
        $modifiedDate,
        $filename);
    }

    return $text;
  }

  /**
   * Formats a number with a bytes suffix.
   *
   * The given number is formatted and returned as markup that may include
   * a suffix like 'K' for kilobytes, 'M' for megabytes, or 'G' for
   * gigabytes.
   *
   * An optional argument selects whether the unit of 'k' should
   * be the current standard 1000, or the older convention of 1024.
   *
   * Optionally, suffixes may be full words instead of abbreviations.
   *
   * @param mixed $number
   *   The number to format.
   * @param int $kunit
   *   (optional, default = 1000) Either 1000 (ISO standard kilobyte) or
   *   1024 (legacy kibibyte).
   * @param bool $fullWord
   *   (optional, default = FALSE) When FALSE, use an abbreviation suffix,
   *   like "KB". When TRUE, use a full word suffix, like "Kilobytes".
   * @param int $decimalDigits
   *   (optional, default = 2) The number of decimal digits for the resulting
   *   number.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   Returns a representation of the number, along with a suffix
   *   that summarizes the value using the selected 'k' units.
   *
   * @internal
   * Drupal's format_size() function is widely used to format values in
   * units of kilobytes, megabytes, etc. Unfortunately, it considers a
   * "kilobyte" equal to 1024 bytes, which is not compliant with the
   * 1998 International Electrotechnical Commission (IEC) and
   * International System of Quantities (ISO) standards. By those standards,
   * a "kilobyte" equals 1000 bytes. This definition retains the SI units
   * convention that "k" is for 1000 (e.g. kilometer, kilogram).
   *
   * The following code is loosely based upon format_size(). It corrects
   * the math for the selected "k" units (kilo or kibi) and uses the
   * corresponding correct suffixes. It also supports a requested number
   * of decimal digits.
   * @endinternal
   */
  public static function formatBytes(
    $number,
    int $kunit = 1000,
    bool $fullWord = FALSE,
    int $decimalDigits = 2) {

    // Validate.
    if ($number <= 0) {
      return "0";
    }
    switch ($kunit) {
      case 1000:
      case 1024:
        $kunit = (float) $kunit;
        break;

      default:
        $kunit = 1000.0;
        break;
    }

    if ($decimalDigits < 0) {
      $decimalDigits = 0;
    }

    if ($number < $kunit) {
      // The quantity is smaller than the 'k' unit. Report it as-is.
      if ($fullWord === TRUE) {
        if ($number === 1) {
          return "1 byte";
        }
        return "$number bytes";
      }

      return ((string) $number) . "B";
    }

    // Find the proper 'k'. Values larger than the largest 'k' unit
    // will just be large numbers ahead of the largest prefix (e.g.
    // "879 YB").
    $value = (((float) $number) / $kunit);
    foreach (['K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'] as $unit) {
      if ($value < $kunit) {
        // The value is now less than the next unit, stop reducing.
        break;
      }

      $value /= $kunit;
    }

    // Simplify the remainder with the proper number of decimal digits.
    $fmt = '%1.' . $decimalDigits . 'f';
    $formattedValue = sprintf($fmt, $value);

    // If using abbrevations, return the formatted value and suffix.
    if ($fullWord === FALSE) {
      // Use abbreviations for suffixes.
      switch ($kunit) {
        default:
        case 1000:
          $suffix = $unit . 'B';
          break;

        case 1024:
          $suffix = $unit . 'iB';
          break;
      }

      return ((string) $formattedValue) . $suffix;
    }

    // When using full words, return the formatted value and word,
    // using singular and plural forms as appropriate.
    $singular = [];
    $plural   = [];
    switch ($kunit) {
      default:
      case 1000:
        $singular = [
          'K' => 'Kilobyte',
          'M' => 'Megabyte',
          'G' => 'Gigabyte',
          'T' => 'Terabyte',
          'P' => 'Petabyte',
          'E' => 'Exabyte',
          'Z' => 'Zettabyte',
          'Y' => 'Yottabyte',
        ];
        $plural = [
          'K' => 'Kilobytes',
          'M' => 'Megabytes',
          'G' => 'Gigabytes',
          'T' => 'Terabytes',
          'P' => 'Petabytes',
          'E' => 'Exabytes',
          'Z' => 'Zettabytes',
          'Y' => 'Yottabytes',
        ];
        break;

      case 1024:
        $singular = [
          'K' => 'Kibibyte',
          'M' => 'Mibibyte',
          'G' => 'Gibibyte',
          'T' => 'Tebibyte',
          'P' => 'Pebibyte',
          'E' => 'Exbibyte',
          'Z' => 'Zebibyte',
          'Y' => 'Yobibyte',
        ];
        $plural = [
          'K' => 'Kibibytes',
          'M' => 'Mibibytes',
          'G' => 'Gibibytes',
          'T' => 'Tebibytes',
          'P' => 'Pebibytes',
          'E' => 'Exbibytes',
          'Z' => 'Zebibytes',
          'Y' => 'Yobibytes',
        ];
        break;
    }

    if ($value <= 1) {
      return ((string) $formattedValue) . $singular[$unit];
    }

    return ((string) $formattedValue) . $plural[$unit];
  }

  /*--------------------------------------------------------------------
   *
   * Usage report formatting.
   *
   *--------------------------------------------------------------------*/

  /**
   * Formats a usage report as text.
   *
   * The keys and values of the array are interpreted as a usage report
   * and formatted as a text table of values.
   *
   * @param array $response
   *   The key-value response to convert to text.
   *
   * @return string
   *   The string representation.
   */
  public static function formatAsUsage(array $response) {
    //
    // Validate
    // --------
    // It is not an error to try and format an empty response array.
    // It could simply mean that there is no response.
    if (empty($response) === TRUE) {
      return '';
    }

    // Make sure the response does appear to be usage information.
    if (isset($response['nFolders']) === FALSE) {
      throw new \InvalidArgumentException(
        "Programmer error: Response array is not in key-value form when formatting for usage data.");
    }

    //
    // Generate output
    // ---------------
    // Pull out user and host information and output that first.
    // Then map the remaining usage information to friendlier names
    // and output that as a key-value list.
    //
    // Get host and user.
    $host = (isset($response['host']) === TRUE) ?
      $response['host'] : 'Unknown host';

    if (isset($response['uid-account-name']) === TRUE) {
      $userName = $response['uid-account-name'];
      if (empty($userName) === TRUE) {
        if (isset($response['uid-display-name']) === TRUE) {
          $userName = $response['uid-display-name'];
        }
      }
    }
    elseif (isset($response['user-account-name']) === TRUE) {
      // Legacy.
      $userName = $response['user-account-name'];
      if (empty($userName) === TRUE) {
        if (isset($response['user-display-name']) === TRUE) {
          $userName = $response['user-display-name'];
        }
      }
    }
    elseif (isset($response['uid']) === TRUE) {
      $userName = $response['uid'];
    }
    elseif (isset($response['user-id']) === TRUE) {
      // Legacy.
      $userName = $response['user-id'];
    }
    else {
      $userName = 'Unknown user';
    }

    $text = "$host usage for $userName:\n";

    // Create an alternate usage array with user-friendly names.
    $alt = [];
    foreach ($response as $k => $r) {
      switch ($k) {
        case 'uid':
        case 'uid-account-name':
        case 'uid-display-name':
        case 'host':
          break;

        case 'user-uid':
        case 'user-account-name':
        case 'user-display-name':
          // Legacy fields that are no longer supported from the server.
          break;

        case 'nFolders':
          $alt['Folders'] = $r;
          break;

        case 'nFiles':
          $alt['Files'] = $r;
          break;

        case 'nBytes':
          $alt['Bytes'] = $r;
          break;

        default:
          $alt[$k] = $r;
          break;
      }
    }

    $text .= self::formatAsText($alt, '  ');
    return $text;
  }

  /*--------------------------------------------------------------------
   *
   * Configuration report formatting.
   *
   *--------------------------------------------------------------------*/

  /**
   * Formats a configuration report as text.
   *
   * The keys and values of the array are interpreted as a configuration report
   * and formatted as a text table of values.
   *
   * @param array $response
   *   The key-value response to convert to text.
   *
   * @return string
   *   The string representation.
   */
  public static function formatAsConfiguration(array $response) {
    //
    // Validate
    // --------
    // It is not an error to try and format an empty response array.
    // It could simply mean that there is no response.
    if (empty($response) === TRUE) {
      return '';
    }

    //
    // Generate output
    // ---------------
    // Show the host first, then map the remaining information into
    // friendlier names and a better structure. Then output that as a
    // key-value list.
    //
    // Get host and user.
    $host = (isset($response['host']) === TRUE) ?
      $response['host'] : 'Unknown host';

    // Create an alternate configuration array with user-friendly names.
    $alt = [];
    $alt['Web services'] = [];
    $alt['File handling'] = [];

    $showExtensions = TRUE;

    foreach ($response as $k => $r) {
      switch ($k) {
        case 'host':
          // Already handled.
          break;

        case 'GET':
        case 'POST':
        case 'DELETE':
        case 'PATCH':
          $webalt = [];
          foreach ($r as $kk => $rr) {
            switch ($kk) {
              case 'serializer-formats':
                $webalt['Formats'] = $rr;
                break;

              case 'authentication-providers':
                $webalt['Authentication'] = $rr;
                break;

              default:
                $webalt[$kk] = $rr;
            }
          }

          $alt['Web services'][$k] = $webalt;
          break;

        case 'file-restrict-extensions':
          $alt['File handling']['Restrict extensions'] = $r;
          if (is_bool($r) === TRUE && $r === FALSE) {
            $showExtensions = FALSE;
          }
          elseif (is_scalar($r) === TRUE &&
              (((string) $r) === "false") || ((string) $r) === "FALSE") {
            $showExtensions = FALSE;
          }
          break;

        case 'file-allowed-extensions':
          $alt['File handling']['Allowed extensions'] = $r;
          break;

        default:
          if (isset($alt['Other']) === FALSE) {
            $alt['Other'] = [];
          }

          $alt['Other'][$k] = $r;
          break;
      }
    }

    if ($showExtensions === FALSE) {
      unset($alt['File handling']['Allowed extensions']);
    }

    $text = "$host configuration:\n";
    $text .= self::formatAsText($alt, '  ');
    return $text;
  }

  /*--------------------------------------------------------------------
   *
   * Version report formatting.
   *
   *--------------------------------------------------------------------*/

  /**
   * Formats a version number report as text.
   *
   * The keys and values of the array are interpreted as a version number
   * report and formatted as a text table of values.
   *
   * @param array $response
   *   The key-value response to convert to text.
   *
   * @return string
   *   The string representation.
   */
  public static function formatAsVersion(array $response) {
    //
    // Validate
    // --------
    // It is not an error to try and format an empty response array.
    // It could simply mean that there is no response. But it is unlikely.
    if (empty($response) === TRUE) {
      return '';
    }

    // Confirm that this does seem to be a version response.
    if (isset($response['client']) === FALSE) {
      throw new \InvalidArgumentException(
        "Programmer error: Response array is not in key-value form when formatting version data.");
    }

    //
    // Generate output
    // ---------------
    // The output has several rows:
    // - The application name and version.
    // - The client API name and version.
    // - The host server name.
    // - A list of server item names and versions.
    //
    // Normally all of these are present in a version request's response.
    // But to be robust we need to check for each one and have a fallback.
    $text = '';
    $indent = '  ';

    //
    // Print versions.
    //
    foreach (['client', 'server'] as $section) {
      if (isset($response[$section]) === FALSE) {
        continue;
      }

      if ($section === 'client') {
        $text .= "Client software\n";
      }
      elseif ($section === 'server') {
        if (isset($response['host']) === FALSE) {
          // If there is no host, then there cannot be server side version
          // numbers. Anything that might be present is bogus. Skip it.
          break;
        }

        $hostname = $response['host'];
        $text .= "\nWeb services at $hostname\n";
      }

      foreach ($response[$section] as $k => $v) {
        $name = $k;
        $version = '';
        $protocol = '';

        if (isset($response[$section][$k]['name']) === TRUE) {
          $name = $response[$section][$k]['name'];
        }

        if (isset($response[$section][$k]['version']) === TRUE) {
          $version = ', version ' . $response[$section][$k]['version'];
        }

        if (isset($response[$section][$k]['protocol_version']) === TRUE) {
          $protocol = ', protocol version ' .
            number_format($response[$section][$k]['protocol_version'], 1);
        }

        $text .= "$indent$name$version$protocol\n";
      }
    }

    return $text;
  }

}
