<?php

/**
 * @file
 * Issues command-line requests to a host using the FolderShare module.
 */

require "FolderShareConnect.php";
require "FolderShareFormat.php";
require "FolderShareCommands.php";

use FolderShare\FolderShareConnect;
use FolderShare\FolderShareFormat;
use FolderShare\FolderShareCommands;

/*----------------------------------------------------------------------
 *
 * Verify execution environment.
 *
 *----------------------------------------------------------------------*/

/**
 * The required PHP version.
 */
define('REQUIRED_PHP_VERSION', '7.0.0');

/**
 * The required PHP packages.
 */
define('REQUIRED_PACKAGES', [
  [
    'extension' => 'curl',
    'name'      => 'PHP CURL',
    'version'   => '7.0.0',
  ],
  [
    'extension' => 'date',
    'name'      => 'PHP Date',
    'version'   => '7.0.0',
  ],
  [
    'extension' => 'json',
    'name'      => 'PHP JSON',
    'version'   => '1.5.0',
  ],
  [
    'extension' => 'mbstring',
    'name'      => 'PHP Multi-byte String',
    'version'   => '7.0.0',
  ],
  [
    'extension' => 'readline',
    'name'      => 'PHP Readline',
    'version'   => '7.0.0',
  ],
]);

/**
 * Verify suitability of PHP installation.
 *
 * This file uses PHP 7 syntax, such as for type hints on function arguments.
 * If the file is read by PHP <7, it will report syntax errors *before* we
 * get a chance to check the version number.
 */
if (version_compare(phpversion(), REQUIRED_PHP_VERSION) < 0) {
  $v = REQUIRED_PHP_VERSION;
  fwrite(STDERR, "Abort. The current version of PHP is out of date.\n");
  fwrite(STDERR, "This application requires at least version $v.\n");
  exit(1);
}

foreach (REQUIRED_PACKAGES as $package) {
  $e = $package['extension'];
  $v = $package['version'];
  $n = $package['name'];
  $pv = phpversion($e);

  if ($pv === FALSE) {
    fwrite(STDERR, "Abort. The required $n package is not installed.\n");
    exit(1);
  }

  if (empty($v) === FALSE && empty($pv) === FALSE) {
    // A specific version number is required.
    if (version_compare($pv, $v) < 0) {
      fwrite(STDERR, "Abort. A newer version of the $n package is required.\n");
      fwrite(STDERR, "This application requires at least version $v, but only $pv is available.\n");
      exit(1);
    }
  }
}

/*----------------------------------------------------------------------
 *
 * Define constants.
 *
 *----------------------------------------------------------------------*/

/**
 * This application name.
 */
define('NAME', 'FolderShare application');

/**
 * This application version number.
 */
define('VERSION', '1.1.0 (February 2020)');

/*--------------------------------------------------------------------
 *
 * Utilities for help and error messages.
 *
 *--------------------------------------------------------------------*/

/**
 * Prints help information for command line use.
 *
 * The printed information is written for users that add commands to
 * the Linux/macOS/BSD shell command line. The host, user name, password,
 * and other base options are listed.
 *
 * @param string $appPath
 *   The path to the application.
 */
function printCommandLineHelp(string $appPath) {

  $appName = basename($appPath);
  $helpStart = <<<EOS
Usage is: $appName [HOST_OPTIONS] [COMMAND] [OPTIONS]

Create, delete, change, and manage files and folders on a remote web site that uses the FolderShare module.

Commands
--------

EOS;

  $helpEnd = <<<EOS

Add "--help" after any command to show its description and flags.

Host and account
----------------
All commands require a remote host and account:

  --host HOSTNAME   Select the host (and optional port) for the site.
  --port NUMBER     Select the port for the site.
  --username NAME   Specify the user login name on the site.
  --password PASS   Specify the user password on the site.

A host name may be a text name or numeric IPv4 or IPv6 address. Text names and IPv4 addresses may include a colon (":") and port number at the end. To include a port number with an IPv6 address, surround the address with square brackets first. Alternatively, use "--port" to specify a port number.

If the port number is omitted, a default is used.

Examples:
  $appName --host example.com --username me --password pw ls /
  $appName --host example.com:80 --username me --password pw ls /
  $appName --host 127.0.0.1 --username me --password pw ls /
  $appName --host 127.0.0.1:80 --username me --password pw ls /
  $appName --host 2001:0db8:85a3:0000:0000:8a2e:0370:7334 --username me --password pw ls /
  $appName --host [2001:0db8:85a3:0000:0000:8a2e:0370:7334]:80 --username me --password pw ls /

If the username and/or password are omitted, the command will prompt for them.

Additional options
------------------
  --help            Show this help information.
  --verbose         Be verbose while performing the operation.
  --version         Show application version numbers.
  --format type     Select how information is displayed:
                      auto     (default) Use an operation-specific default
                      full     Show full-detail nested-array output
                      keyvalue Show a table of key-value pairs
                      linux    Show a Linux-style output
                      text     Show key-value pairs as text

For Linux/macOS/BSD-style commands, the format type defaults to 'linux' and output is formatted similar to those OS commands.

Example:
  $appName --host myhost --format full stat /myfolder

Paths
-----
Remote file and folder paths have the form:

  [SCHEME:][//USER]/PATH.

SCHEME is optional and one of:
  "personal"  (default) Your files and folders, and those shared with you.
  "public"    The host's public files and folders available to anyone.

USER is optional and selects the user ID or account name. It defaults to the current user.

PATH is a file and folder path, separated by forward slashes. Wildcards are not accepted. If a path contains spaces or special characters, the whole path should be included in quotes.

Examples:
  $appName --host example.com ls /
  $appName --host example.com ls personal:/
  $appName --host example.com ls public:/

EOS;

  $help = $helpStart;

  $commandNames = array_keys(FolderShareCommands::COMMANDS);
  sort($commandNames, (SORT_STRING | SORT_FLAG_CASE));

  foreach ($commandNames as $commandName) {
    $commandInfo = FolderShareCommands::COMMANDS[$commandName];

    if (isset($commandInfo['brief']) === TRUE) {
      $brief = $commandInfo['brief'];
    }
    else {
      $brief = $commandName;
    }

    $help .= sprintf("  %-18s%s\n", $commandName, $brief);
  }

  $help .= $helpEnd;

  fwrite(STDOUT, $help);
}

/**
 * Prints help information for prompt use.
 *
 * The printed information is written for users that are using the interactive
 * prompt. The host, user name, password, and other base options are not listed
 * since they would already have been provided.
 *
 * @param string $appPath
 *   The path to the application.
 */
function printPromptHelp(string $appPath) {
  $helpStart = <<<EOS
Commands
--------

EOS;

  $helpEnd = <<<EOS

Add "--help" after any command to show its description and flags.

Paths
-----
Remote file and folder paths have the form:

  [SCHEME:][//USER]/PATH.

SCHEME is optional and one of:
  "personal"  (default) Your files and folders, and those shared with you.
  "public"    The host's public files and folders available to anyone.

USER is optional and selects the user ID or account name. It defaults to the current user.

PATH is a file and folder path, separated by forward slashes. Wildcards are not accepted. If a path contains spaces or special characters, the whole path should be included in quotes.

Examples:
  ls /
  ls personal:/
  ls public:/

EOS;

  $help = $helpStart;

  $commandNames = array_keys(FolderShareCommands::COMMANDS);
  $commandNames[] = 'help';
  $commandNames[] = 'quit';
  $commandNames[] = 'exit';
  sort($commandNames, (SORT_STRING | SORT_FLAG_CASE));

  foreach ($commandNames as $commandName) {
    if ($commandName === 'help') {
      $brief = "Show this help information.";
    }
    elseif ($commandName === 'quit') {
      $brief = "Quit.";
    }
    elseif ($commandName === 'exit') {
      $brief = "Quit.";
    }
    else {
      $commandInfo = FolderShareCommands::COMMANDS[$commandName];

      if (isset($commandInfo['brief']) === TRUE) {
        $brief = $commandInfo['brief'];
      }
      else {
        $brief = $commandName;
      }
    }

    $help .= sprintf("  %-18s%s\n", $commandName, $brief);
  }

  $help .= $helpEnd;

  fwrite(STDOUT, $help);
}

/**
 * Prints help information for a command.
 *
 * @param string $appName
 *   The name of this command-line application.
 * @param string $commandName
 *   The name of the command.
 */
function printCommandHelp(string $appName, string $commandName) {
  if (isset(FolderShareCommands::COMMANDS[$commandName]) === FALSE) {
    // Unknown command. Print nothing.
    return;
  }

  $command = FolderShareCommands::COMMANDS[$commandName];

  //
  // Print synopsis
  // --------------
  // Print a command-line template for the command.
  $synopsis = "Usage: $appName [HOST_OPTIONS] ";
  if (isset($command['synopsis']) === FALSE) {
    $synopsis .= $commandName;
  }
  else {
    $synopsis .= $command['synopsis'];
  }

  fwrite(STDOUT, "$synopsis\n");

  //
  // Print description
  // -----------------
  // Print a multi-line text description.
  $description = "\n";
  if (isset($command['description']) === TRUE) {
    $description .= $command['description'] . "\n";
  }

  fwrite(STDOUT, $description);

  //
  // Print flags (if any)
  // -----------
  // Print a list of flags and their descriptions.
  if (isset($command['flags']) === TRUE) {
    fwrite(STDOUT, "\nOptions:\n");

    // Get flag names then sort them.
    $flagNames = array_keys($command['flags']);
    sort($flagNames, (SORT_STRING | SORT_FLAG_CASE));

    // Print them out.
    $flagIndent = '  ';
    $flagHelpIndent = '      ';

    foreach ($flagNames as $flagName) {
      $flagInfo = &$command['flags'][$flagName];

      // Print the flag line, with synonyms.
      $flagLine = "$flagIndent$flagName";
      if (isset($flagInfo['synonyms']) === TRUE) {
        foreach ($flagInfo['synonyms'] as $flagSynonym) {
          $flagLine .= ", $flagSynonym";
        }
      }

      fwrite(STDOUT, "$flagLine\n");

      // Print the flag command.
      if (isset($flagInfo['brief']) === TRUE) {
        $helpLines = mb_split("\n", $flagInfo['brief']);
        foreach ($helpLines as $helpLine) {
          fwrite(STDOUT, "$flagHelpIndent$helpLine\n");
        }
      }
    }
  }

  //
  // Print command synonyms (if any)
  // ----------------------
  // Print a list of synonyms.
  if (isset($command['synonyms']) === TRUE) {
    $names = [$commandName];
    foreach ($command['synonyms'] as $synonym) {
      $names[] = $synonym;
    }

    sort($names, SORT_NATURAL);
    $text = implode(', ', $names);

    fwrite(STDOUT, "\nSynonyms:\n  $text\n");
  }
}

/**
 * Prints the client side version to STDERR and exits.
 *
 * @param array $options
 *   The command-line options array.
 */
function printVersionAndExit(array $options) {
  // Create a dummy server connection without a host or authentication.
  // Then get the version of the client side only.
  $dummyServer = new FolderShareConnect();
  $options['format'] = 'text';
  $text = FolderShareCommands::version($dummyServer, $options, FALSE);
  fwrite(STDERR, $text);
  exit(1);
}

/**
 * Prints an error message to STDERR and exits.
 *
 * @param string $appPath
 *   The path to the application.
 * @param string $message
 *   The error message.
 */
function printErrorAndExit(string $appPath, string $message) {
  $appName = basename($appPath, '.php');
  fwrite(STDERR, "$appName: $message\n");
  exit(1);
}

/*--------------------------------------------------------------------
 *
 * Command-line parsing.
 *
 * These functions parse the command line.
 *
 *--------------------------------------------------------------------*/

/**
 * Parses the command line and returns values.
 *
 * The command line is parsed and its values validated and returned in
 * an associative array.
 *
 * @param array $argv
 *   The array of command-line arguments.
 */
function parseCommandLine(array $argv) {
  //
  // Set initial values
  // ------------------
  // Get the application name and set up defaults for all values.
  $options = [];
  $appPath = $options['appPath'] = array_shift($argv);
  $appName = $options['appName'] = basename($appPath, '.php');

  // Connection options.
  $options['host'] = '';
  $options['port'] = '';
  $options['username'] = '';
  $options['password'] = '';

  // Flags.
  $options['format'] = 'auto';
  $options['verbose'] = FALSE;

  // Command.
  $options['command'] = [];
  $options['paths'] = [];
  $options['flags'] = [];

  $showVersion = FALSE;

  $usage = "\nUse '--help' for a list of commands and options.";

  //
  // Parse arguments
  // ---------------
  // Loop through the arguments and look for options and commands.
  // When a command name is found, all further arguments are considered
  // part of that command.
  $n = count($argv);
  for ($i = 0; $i < $n; ++$i) {
    // The first argument that does not start with '-' is the command.
    // Add it to the command's argument list, along with all further
    // arguments. It is up to the command to parse them.
    $dash = mb_substr($argv[$i], 0, 1);
    if ($dash !== '-') {
      // Command found!
      $options['command'] = array_slice($argv, $i);
      break;
    }

    // Otherwise, parse generic options relating to the connection
    // or flags.
    //
    // Ignore one or two leading dashes while parsing options.
    $dashes = mb_substr($argv[$i], 0, 2);
    if ($dashes === '--') {
      $arg = mb_substr($argv[$i], 2);
    }
    else {
      $arg = mb_substr($argv[$i], 1);
    }

    switch ($arg) {
      // Immediate response options.
      case 'help':
        // Print help information and stop.
        printCommandLineHelp($appName);
        exit(0);

      case 'version':
        // Print application version and stop.
        $showVersion = TRUE;
        break;

      // Connection options.
      case 'host':
        if (($i + 1) >= $n) {
          printErrorAndExit(
            $appName,
            "A host name is required after '--$arg'.$usage");
        }

        ++$i;
        $options['host'] = $argv[$i];
        break;

      case 'port':
        if (($i + 1) >= $n) {
          printErrorAndExit(
            $appName,
            "A port number is required after '--$arg'.$usage");
        }

        ++$i;
        $options['port'] = $argv[$i];
        break;

      case 'user':
      case 'username':
        if (($i + 1) >= $n) {
          printErrorAndExit(
            $appName,
            "A user name is required after '--$arg'.$usage");
        }

        ++$i;
        $options['username'] = $argv[$i];
        break;

      case 'password':
        if (($i + 1) >= $n) {
          printErrorAndExit(
            $appName,
            "A password is required after '--$arg'.$usage");
        }

        ++$i;
        $options['password'] = $argv[$i];
        break;

      // Flags.
      case 'verbose':
        $options['verbose'] = TRUE;
        break;

      case 'format':
        if (($i + 1) >= $n) {
          printErrorAndExit(
            $appName,
            "A format name is required after '--$arg'.$usage");
        }

        ++$i;
        $value = $argv[$i];
        switch ($value) {
          // Principal values.
          case 'auto':
          case 'full':
          case 'keyvalue':
          case 'linux':
          case 'text':
            break;

          // Synonyms.
          case 'all':
          case 'raw':
            $value = 'full';
            break;

          case 'kv':
            $value = 'keyvalue';
            break;

          case 'macos':
          case 'osx':
          case 'unix':
          case 'bsd':
            $value = 'linux';
            break;

          case 'default':
            $value = 'auto';
            break;

          default:
            printErrorAndExit(
              $appName,
              "Unknown format name: '$value'.$usage");
            break;
        }

        $options['format'] = $value;
        break;

      default:
        // Otherwise the option is not generic.
        $a = $argv[$i];
        printErrorAndExit(
          $appName,
          "Unknown option: '--$a'.$usage");
        break;
    }
  }

  //
  // Handle version
  // --------------
  // The --version flag can be used with or without a host.
  //
  // Without a host, just show the client side's versions.
  //
  // With a host, map --version to the 'version' command.
  if ($showVersion === TRUE) {
    if (empty($options['host']) === TRUE) {
      printVersionAndExit($options);
    }

    // Override any additional command given and make it "version".
    $options['command'] = ["version"];
  }

  //
  // Validate
  // --------
  // Check that a host name has been given.
  if (empty($options['host']) === TRUE) {
    printErrorAndExit(
      $appName,
      "A host name is required. Use '--host hostname'.$usage");
  }

  return $options;
}

/**
 * Validates a command and its flags.
 *
 * @param array $options
 *   The command-line options array.
 * @param bool $exitOnError
 *   (optional, default = TRUE) When TRUE, on an error a message is output
 *   and the application exits. When FALSE, on an error a message is output
 *   and FALSE is returned.
 *
 * @return bool
 *   Returns TRUE on successful validation, and FALSE otherwise.
 */
function validateCommand(array &$options, bool $exitOnError = TRUE) {
  //
  // Recognize command
  // -----------------
  // Check if the command is recognized. It can be a primary command name
  // or one of the synonyms.
  $appName = $options['appName'];
  if ($exitOnError === TRUE) {
    $usage = "\nUse '--help' for a list of commands and options.";
  }
  else {
    $usage = "\nType 'help' for a list of commands and options.\n";
  }

  if (isset($options['command']) === FALSE) {
    if ($exitOnError === TRUE) {
      printErrorAndExit(
        $appName,
        "Missing command.$usage");
    }

    print("Missing command.$usage");
    return FALSE;
  }

  $commandName = reset($options['command']);
  $found = FALSE;

  if (isset(FolderShareCommands::COMMANDS[$commandName]) === TRUE) {
    $found = TRUE;
  }
  else {
    // Look through the synonyms of all commands. If found, swap the
    // command name from the synonym to the primary name.
    foreach (FolderShareCommands::COMMANDS as $name => &$info) {
      if (isset($info['synonyms']) === TRUE) {
        if (in_array($commandName, $info['synonyms']) === TRUE) {
          $found = TRUE;
          $commandName = $name;
          $options['command'][0] = $name;
          break;
        }
      }
    }
  }

  if ($found === FALSE) {
    if ($exitOnError === TRUE) {
      printErrorAndExit(
        $appName,
        "Unknown command: '$commandName'.$usage");
    }

    print("Unknown command: '$commandName'.$usage");
    return FALSE;
  }

  //
  // Check implementation
  // --------------------
  // Make sure the command has a function! If it doesn't, the command is
  // not yet implemented.
  if (isset(FolderShareCommands::COMMANDS[$commandName]['function']) === FALSE) {
    if ($exitOnError === TRUE) {
      printErrorAndExit(
        $appName,
        "$commandName is not supported yet.");
    }

    print("$commandName is not supported yet.\n");
    return FALSE;
  }

  //
  // Handle return formats
  // ---------------------
  // Map the 'auto' return type to the actual type.
  if ($options['format'] === 'auto') {
    $options['format'] = FolderShareCommands::COMMANDS[$commandName]['defaultFormat'];
  }

  // Check that only valid return types were provided.
  if (empty(FolderShareCommands::COMMANDS[$commandName]['formats']) === FALSE &&
      in_array($options['format'], FolderShareCommands::COMMANDS[$commandName]['formats']) === FALSE) {
    $rt = $options['format'];
    if ($exitOnError === TRUE) {
      printErrorAndExit(
        $appName,
        "Unsupported format name '$rt' for '$commandName'.$usage");
    }

    print("Unsupported format name '$rt' for '$commandName'.$usage");
    return FALSE;
  }

  //
  // Parse flags and paths
  // ---------------------
  // Check that the flags make sense.
  $pathsAndFlags = parseCommandFlags(
    $appName,
    $commandName,
    $options['command']);
  if ($pathsAndFlags === FALSE) {
    return FALSE;
  }

  $options['paths'] = $pathsAndFlags['paths'];
  $options['flags'] = $pathsAndFlags['flags'];

  return TRUE;
}

/**
 * Parses flags for a specific command.
 *
 * An array arguments for a command is split into:
 * - Flags that start with '-' or '--'.
 * - Paths or other non-dash arguments.
 *
 * Flags are parsed for the command and errors generated for unrecognized
 * flags.
 *
 * Flags with a single dash are presumed to be single-letter flags
 * (e.g. "-a"). If multiple letters are given, the flags are expanded
 * into multiple flags (e.g. "-abc" becomes "-a", "-b", and "-c).
 *
 * Flags with a double dash are presumed to be multi-letter word flags
 * and are not changed.
 *
 * Flag synonyms are transparently mapped to their primary names (e.g.
 * "--stuff" becomes "-s" if "-s" is the primary name).
 *
 * @param string $appName
 *   The name of the command-line application.
 * @param string $commandName
 *   The name of the command.
 * @param array $args
 *   The arguments array from the basic command-line parser.
 *
 * @return array|bool
 *   Returns an associative array with 'flags' and 'paths' keys whose
 *   values are each arrays. The 'paths' array contains a list of
 *   paths or non-flag arguments, in order. The 'flags' array contains
 *   a list of flags. Returns FALSE on error.
 */
function parseCommandFlags(string $appName, string $commandName, array $args) {
  //
  // Split args
  // ----------
  // Divide up the incoming arguments into paths and flags. A flag starts
  // with a '-' or '--', while a path does not.
  $paths = [];
  $flags = [];

  array_shift($args);
  foreach ($args as $arg) {
    // Look at the first character to decide if the argument is a flag
    // or a path.
    $dash = mb_substr($arg, 0, 1);
    if ($dash === '-') {
      $flags[] = $arg;
    }
    else {
      $paths[] = $arg;
    }
  }

  //
  // Expand flags
  // ------------
  // For single-dash options with multiple letters, split the letters into
  // separate flags. For instance "-abc" becomes "-a", "-b", and "-c".
  // Leave double-dash options alone.
  $expandedFlags = [];
  foreach ($flags as $flag) {
    // Don't expand double-dash flags.
    $firstTwo = mb_substr($flag, 0, 2);
    if ($firstTwo === '--') {
      $expandedFlags[] = $flag;
      continue;
    }

    // No need to expand flags that are one letter after '-'.
    $len = mb_strlen($flag);
    if ($len === 2) {
      $expandedFlags[] = $flag;
      continue;
    }

    for ($i = 1; $i < $len; ++$i) {
      $f = mb_substr($flag, $i, 1);
      $expandedFlags[] = '-' . $f;
    }
  }

  $flags = $expandedFlags;
  unset($expandedFlags);

  //
  // Simplify flags
  // --------------
  // For all flags, map synonyms to their primary flag name.
  $flagsInfo = FolderShareCommands::COMMANDS[$commandName]['flags'];

  $simplifiedFlags = [];
  foreach ($flags as $flag) {
    if (isset($flagsInfo[$flag]) === TRUE) {
      // The flag is a primary flag name.
      $simplifiedFlags[] = $flag;
    }
    else {
      $found = FALSE;
      foreach ($flagsInfo as $flagName => &$flagInfo) {
        if (isset($flagInfo['synonyms']) === TRUE) {
          foreach ($flagInfo['synonyms'] as $synonymName) {
            if ($flag === $synonymName) {
              // Map the synonym to the primary flag name.
              $simplifiedFlags[] = $flagName;
              $found = TRUE;
              break 2;
            }
          }
        }
      }

      // If the flag was not found as a synonym of anything, then
      // report an error.
      if ($found === FALSE) {
        print("$commandName: Unrecognized option \"$flag\".\n");
        return FALSE;
      }
    }
  }

  return [
    'paths' => $paths,
    'flags' => $simplifiedFlags,
  ];
}

/*--------------------------------------------------------------------
 *
 * Execute!
 *
 * Run the application.
 *
 *--------------------------------------------------------------------*/

//
// Set unlimited memory use. File uploads and downloads can require
// large memory buffers, and the larger the file the larger the buffer.
// To avoid hitting a memory limit, disable it.
//
ini_set('memory_limit', (-1));

//
// Set unlimited execution time. PHP's command-line scripts are supposed to
// default to unlimited execution time, but be sure. This avoids time limit
// interrupts during long file transfers.
//
set_time_limit(0);

//
// Parse the command line. There may not be a command.
//
$options = parseCommandLine($argv);
$appName = $options['appName'];

//
// Open server connection
// ----------------------
// If no user name was provided, assume anonymous access to the site.
// If a user name was given, but no password, prompt for the password.
//
// Then use the given host and open a server connection and set the return
// format and verbosity level.
try {
  if (empty($options['username']) === FALSE &&
      $options['username'] !== FolderShareCommands::ANONYMOUS) {
    if (empty($options['password']) === TRUE) {
      // Prompt for password.
      print("Password: ");
      $options['password'] = stream_get_line(STDIN, 1024, PHP_EOL);
    }
  }
  elseif ($options['username'] === FolderShareCommands::ANONYMOUS) {
    $options['username'] = '';
  }

  // Open server connection.
  $server = new FolderShareConnect();

  // Enable/disable verbosity.
  if ($options['verbose'] === TRUE) {
    $server->setVerbose(TRUE);
  }

  // Set the host and port.
  $server->setHostNameAndPortNumber($options['host'], $options['port']);

  // Log in.
  //
  // Provide the user name and password to log in. This typically triggers
  // a connection. It will fail if the host, user name, or password are bad,
  // or if there are problems communicating with the host.
  if ($server->login($options['username'], $options['password']) === FALSE) {
    printErrorAndExit($appName, "Login failed.");
  }
}
catch (\Exception $e) {
  printErrorAndExit($appName, $e->getMessage());
}

//
// Execute single command
// ----------------------
// If a command was given on the command line, execute it.
if (empty($options['command']) === FALSE) {
  //
  // Validate command
  // ----------------
  // If a command is given, it must be valid.
  // - Check if the command is a primary name or a synonym.
  // - Check that there is a function for the command.
  // - Check that the return format is supported.
  // - Check that the given flags make sense.
  if (empty($options['command']) === FALSE) {
    validateCommand($options);
  }

  if (in_array('--help', $options['flags']) === TRUE) {
    printCommandHelp($appName, $options['command'][0]);
    exit(0);
  }

  // Set the preferred return format.
  if ($options['format'] === 'linux' || $options['format'] === 'text') {
    $server->setReturnFormat('keyvalue');
  }
  else {
    $server->setReturnFormat($options['format']);
  }

  try {
    //
    // Dispatch
    // --------
    // Invoke the command.
    $commandName = reset($options['command']);
    $fun = FolderShareCommands::COMMANDS[$commandName]['function'];
    if (is_callable($fun) === FALSE) {
      $fun = 'FolderShare\\' . $fun;
      if (is_callable($fun) === FALSE) {
        printErrorAndExit($appName, "Cannot find function: $fun");
      }
    }
    $status = $fun($server, $options);
    exit(($status === TRUE) ? 0 : 1);
  }
  catch (\Exception $e) {
    printErrorAndExit($appName, $e->getMessage());
  }
}

//
// Execute multiple commands
// -------------------------
// Enter a prompt loop and execute commands as given.
$username = $server->getUserName();
if (empty($username) === TRUE) {
  $username = FolderShareCommands::ANONYMOUS;
}

printf("Foldershare shell.\nLogged in as '%s'.\nType 'help' for a list of commands, 'quit' or 'exit' to exit.\n", $username);

while (TRUE) {
  try {
    //
    // Prompt
    // ------
    // Get a single line of input.
    $prompt = "$appName> ";
    if (PHP_OS === 'WINNT') {
      $line = stream_get_line(STDIN, 1024, PHP_EOL);
    }
    else {
      $line = readline($prompt);
    }

    if ($line === FALSE) {
      // EOF.
      break;
    }

    if (empty($line) === TRUE) {
      continue;
    }

    if (PHP_OS !== 'WINNT') {
      readline_add_history($line);
    }

    //
    // Split into arguments
    // --------------------
    // Quotes require special handling to keep values together as a single
    // argument.
    $parts = mb_split('["\']', $line);
    $inQuoted = FALSE;
    $args = [];
    foreach ($parts as $part) {
      if ($inQuoted === TRUE) {
        // This part is a quoted string, with the quotes removed. Do not
        // trim or change in any way.
        $args[] = $part;
        $inQuoted = FALSE;
      }
      else {
        // The part is one or more non-quoted arguments. Remove redundant
        // white space, then split into arguments at white space boundaries.
        foreach (mb_split(' ', mb_ereg_replace('\s+', ' ', $part)) as $p) {
          if (empty($p) === FALSE) {
            $args[] = $p;
          }
        }

        $inQuoted = TRUE;
      }
    }

    // Build an options array.
    $localOptions = $options;
    $localOptions['command'] = $args;

    $commandName = reset($localOptions['command']);
    switch ($commandName) {
      case 'help':
        printPromptHelp($appName);
        continue 2;

      case 'quit':
      case 'exit':
        exit(0);
    }

    //
    // Validate command
    // ----------------
    // If a command is given, it must be valid.
    // - Check if the command is a primary name or a synonym.
    // - Check that there is a function for the command.
    // - Check that the return format is supported.
    // - Check that the given flags make sense.
    if (validateCommand($localOptions, FALSE) === FALSE) {
      continue;
    }

    $commandName = reset($localOptions['command']);

    if (in_array('--help', $localOptions['flags']) === TRUE) {
      printCommandHelp($appName, $commandName);
      continue;
    }

    // Set the preferred return format.
    if ($localOptions['format'] === 'linux' ||
        $localOptions['format'] === 'text') {
      $server->setReturnFormat('keyvalue');
    }
    else {
      $server->setReturnFormat($localOptions['format']);
    }

    //
    // Dispatch
    // --------
    // Invoke the command.
    FolderShareCommands::COMMANDS[$commandName]['function']($server, $localOptions);
  }
  catch (\Exception $e) {
    print($e->getMessage());
  }
}
